# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import difflib
from collections.abc import Callable, Mapping
from typing import TYPE_CHECKING, TypeAlias, overload

from ... import utils as ut
from ...localize import _
from ...utils import ExprProxy
from ..mapdata import (
    GetLocationIndex,
    GetStringIndex,
    GetSwitchIndex,
    GetUnitIndex,
)
from .consttype import ConstType, T, U, _Byte, _Dword, _Word
from .strdict import (
    DefAIScriptDict,
    DefButtonSetDict,
    DefFlingyDict,
    DefIconDict,
    DefImageDict,
    DefIscriptDict,
    DefLocationDict,
    DefPortraitDict,
    DefSpriteDict,
    DefStatTextDict,
    DefSwitchDict,
    DefTechDict,
    DefUnitDict,
    DefUnitOrderDict,
    DefUpgradeDict,
    DefWeaponDict,
)

if TYPE_CHECKING:
    from ..allocator import ConstExpr
    from ..variable import EUDVariable


class TrgAIScript(ConstType):
    __slots__ = ()

    @classmethod
    def cast(cls, s):
        return EncodeAIScript(s)


class TrgLocation(ConstType):
    __slots__ = ()

    @classmethod
    def cast(cls, s):
        return EncodeLocation(s)


class TrgString(ConstType):
    __slots__ = ()

    @classmethod
    def cast(cls, s):
        return EncodeString(s)


class TrgSwitch(ConstType):
    __slots__ = ()

    @classmethod
    def cast(cls, s):
        return EncodeSwitch(s)


class Iscript(ConstType):
    __slots__ = ()

    @classmethod
    def cast(cls, s):
        return EncodeIscript(s)


class StatText(ConstType):
    __slots__ = ()
    _encoding = "UTF-8"
    _data = b""

    @classmethod
    def cast(cls, s):
        return EncodeTBL(s)


class Icon(ConstType):
    __slots__ = ()

    @classmethod
    def cast(cls, s):
        return EncodeIcon(s)


class Portrait(ConstType):
    __slots__ = ()

    @classmethod
    def cast(cls, s):
        return EncodePortrait(s)


# argument types
_ExprProxy: TypeAlias = (
    "ExprProxy[str | bytes | int | EUDVariable | ConstExpr | ExprProxy]"
)
_Arg: TypeAlias = "str | bytes | int | EUDVariable | ConstExpr | ExprProxy[str | bytes | int | EUDVariable | ConstExpr | ExprProxy]"  # noqa: E501
__ExprProxy: TypeAlias = "ExprProxy[str | bytes | int | EUDVariable | ExprProxy]"
__Arg: TypeAlias = "str | bytes | int | EUDVariable | ExprProxy[str | bytes | int | EUDVariable | ExprProxy]"  # noqa: E501


@overload
def EncodeAIScript(ais: str | bytes) -> int:
    ...


@overload
def EncodeAIScript(ais: T) -> T:
    ...


@overload
def EncodeAIScript(ais: _ExprProxy) -> _Dword:
    ...


def EncodeAIScript(ais: _Arg) -> _Dword:  # noqa: N802
    ai = ut.unProxy(ais)

    if isinstance(ai, str):
        ai = ut.u2b(ai)

    if isinstance(ai, bytes):
        ut.ep_assert(len(ai) >= 4, _("AIScript name too short"))

        if len(ai) > 4:
            if ai in DefAIScriptDict:
                return ut.b2i4(DefAIScriptDict[ai])
            sl = _("Cannot encode string {} as {}.").format(ai, "AIScript")
            for match in difflib.get_close_matches(ai, DefAIScriptDict.keys()):
                sl += "\n" + _(" - Suggestion: {}").format(match)

        elif len(ai) == 4:
            if ai in DefAIScriptDict.values():
                return ut.b2i4(ai)
            sl = _("Cannot encode string {} as {}.").format(ai, "AIScript")
            for match in difflib.get_close_matches(ai, DefAIScriptDict.values()):
                sl += "\n" + _(" - Suggestion: {}").format(match)
        raise ut.EPError(sl)

    if isinstance(ai, ConstType):
        raise ut.EPError(_('"{}" is not a {}').format(ais, "AIScript"))
    assert not isinstance(ai, ExprProxy), "unreachable"
    return ai


@overload
def _EncodeAny(t: str, f: Callable, dl: Mapping[str, int], s: str | bytes) -> int:
    ...


@overload
def _EncodeAny(t: str, f: Callable, dl: Mapping[str, int], s: T) -> T:
    ...


@overload
def _EncodeAny(t: str, f: Callable, dl: Mapping[str, int], s: _ExprProxy) -> _Dword:
    ...


def _EncodeAny(t: str, f: Callable, dl: Mapping[str, int], s: _Arg) -> _Dword:  # noqa: N802
    if isinstance(s, int):
        return s

    u = ut.unProxy(s)

    if isinstance(u, (str, bytes)):  # noqa: UP038
        try:
            return f(u)
        except KeyError:
            if isinstance(u, str):
                if u in dl:
                    return dl[u]
                sl = _("Cannot encode string {} as {}.").format(u, t)
                for match in difflib.get_close_matches(u, dl.keys()):
                    sl += "\n" + _(" - Suggestion: {}").format(match)
                raise ut.EPError(sl)
            raise ut.EPError(_('"{}" is not a {}').format(u, t))

    elif isinstance(u, ConstType):
        raise ut.EPError(_('"{}" is not a {}').format(u, t))
    assert not isinstance(u, ExprProxy), "unreachable"
    return u


@overload
def EncodeLocation(loc: str | bytes) -> int:
    ...


@overload
def EncodeLocation(loc: U) -> U:
    ...


@overload
def EncodeLocation(loc: __ExprProxy) -> _Word:
    ...


def EncodeLocation(loc: __Arg) -> _Word:  # noqa: N802
    return _EncodeAny("location", GetLocationIndex, DefLocationDict, loc)  # type: ignore[return-value]


@overload
def EncodeString(s: str | bytes) -> int:
    ...


@overload
def EncodeString(s: U) -> U:
    ...


@overload
def EncodeString(s: __ExprProxy) -> _Word:
    ...


def EncodeString(s: __Arg) -> _Word:  # noqa: N802
    _def_string_dict: dict[str, int] = {}
    return _EncodeAny("MapString", GetStringIndex, _def_string_dict, s)  # type: ignore[return-value]


@overload
def EncodeSwitch(sw: str | bytes) -> int:
    ...


@overload
def EncodeSwitch(sw: U) -> U:
    ...


@overload
def EncodeSwitch(sw: __ExprProxy) -> _Byte:
    ...


def EncodeSwitch(sw: __Arg) -> _Byte:  # noqa: N802
    return _EncodeAny("switch", GetSwitchIndex, DefSwitchDict, sw)  # type: ignore[return-value]


@overload
def EncodeUnit(u: str | bytes) -> int:
    ...


@overload
def EncodeUnit(u: U) -> U:
    ...


@overload
def EncodeUnit(u: __ExprProxy) -> _Word:
    ...


def EncodeUnit(u: __Arg) -> _Word:  # noqa: N802
    return _EncodeAny("unit", GetUnitIndex, DefUnitDict, u)  # type: ignore[return-value]


@overload
def EncodeTBL(t: str | bytes) -> int:
    ...


@overload
def EncodeTBL(t: T) -> T:
    ...


@overload
def EncodeTBL(t: _ExprProxy) -> _Dword:
    ...


def EncodeTBL(t: _Arg) -> _Dword:  # noqa: N802
    # TODO: handle custom stat_txt.tbl
    return _EncodeAny("stat_txt.tbl", lambda s: {}[s], DefStatTextDict, t)


@overload
def EncodeFlingy(flingy: str | bytes) -> int:
    ...


@overload
def EncodeFlingy(flingy: U) -> U:
    ...


@overload
def EncodeFlingy(flingy: __ExprProxy) -> _Word:
    ...


def EncodeFlingy(flingy: __Arg) -> _Word:  # noqa: N802
    return _EncodeAny("flingy", lambda s: {}[s], DefFlingyDict, flingy)  # type: ignore[return-value]


@overload
def EncodeIcon(icon: str | bytes) -> int:
    ...


@overload
def EncodeIcon(icon: U) -> U:
    ...


@overload
def EncodeIcon(icon: __ExprProxy) -> _Word:
    ...


def EncodeIcon(icon: __Arg) -> _Word:  # noqa: N802
    return _EncodeAny("icon", lambda s: {}[s], DefIconDict, icon)  # type: ignore[return-value]


@overload
def EncodeSprite(sprite: str | bytes) -> int:
    ...


@overload
def EncodeSprite(sprite: U) -> U:
    ...


@overload
def EncodeSprite(sprite: __ExprProxy) -> _Word:
    ...


def EncodeSprite(sprite: __Arg) -> _Word:  # noqa: N802
    return _EncodeAny("sprite", lambda s: {}[s], DefSpriteDict, sprite)  # type: ignore[return-value]


@overload
def EncodeImage(image: str | bytes) -> int:
    ...


@overload
def EncodeImage(image: U) -> U:
    ...


@overload
def EncodeImage(image: __ExprProxy) -> _Word:
    ...


def EncodeImage(image: __Arg) -> _Word:  # noqa: N802
    return _EncodeAny("image", lambda s: {}[s], DefImageDict, image)  # type: ignore[return-value]


@overload
def EncodeIscript(iscript: str | bytes) -> int:
    ...


@overload
def EncodeIscript(iscript: U) -> U:
    ...


@overload
def EncodeIscript(iscript: __ExprProxy) -> _Word:
    ...


def EncodeIscript(iscript: __Arg) -> _Word:  # noqa: N802
    return _EncodeAny("iscript", lambda s: {}[s], DefIscriptDict, iscript)  # type: ignore[return-value]


@overload
def EncodeUnitOrder(order: str | bytes) -> int:
    ...


@overload
def EncodeUnitOrder(order: U) -> U:
    ...


@overload
def EncodeUnitOrder(order: __ExprProxy) -> _Byte:
    ...


def EncodeUnitOrder(order: __Arg) -> _Byte:  # noqa: N802
    return _EncodeAny("UnitOrder", lambda s: {}[s], DefUnitOrderDict, order)  # type: ignore[return-value]


@overload
def EncodeWeapon(weapon: str | bytes) -> int:
    ...


@overload
def EncodeWeapon(weapon: U) -> U:
    ...


@overload
def EncodeWeapon(weapon: __ExprProxy) -> _Byte:
    ...


def EncodeWeapon(weapon: __Arg) -> _Byte:  # noqa: N802
    return _EncodeAny("weapon", lambda s: {}[s], DefWeaponDict, weapon)  # type: ignore[return-value]


@overload
def EncodeTech(tech: str | bytes) -> int:
    ...


@overload
def EncodeTech(tech: U) -> U:
    ...


@overload
def EncodeTech(tech: __ExprProxy) -> _Byte:
    ...


def EncodeTech(tech: __Arg) -> _Byte:  # noqa: N802
    return _EncodeAny("tech", lambda s: {}[s], DefTechDict, tech)  # type: ignore[return-value]


@overload
def EncodeUpgrade(upgrade: str | bytes) -> int:
    ...


@overload
def EncodeUpgrade(upgrade: U) -> U:
    ...


@overload
def EncodeUpgrade(upgrade: __ExprProxy) -> _Byte:
    ...


def EncodeUpgrade(upgrade: __Arg) -> _Byte:  # noqa: N802
    return _EncodeAny("upgrade", lambda s: {}[s], DefUpgradeDict, upgrade)  # type: ignore[return-value]


@overload
def EncodePortrait(portrait: str | bytes) -> int:
    ...


@overload
def EncodePortrait(portrait: U) -> U:
    ...


@overload
def EncodePortrait(portrait: __ExprProxy) -> _Word:
    ...


def EncodePortrait(portrait: __Arg) -> _Word:  # noqa: N802
    return _EncodeAny("portrait", lambda s: {}[s], DefPortraitDict, portrait)  # type: ignore[return-value]


@overload
def EncodeButtonSet(buttonset: str | bytes) -> int:
    ...


@overload
def EncodeButtonSet(buttonset: U) -> U:
    ...


@overload
def EncodeButtonSet(buttonset: __ExprProxy) -> _Word:
    ...


def EncodeButtonSet(buttonset: __Arg) -> _Word:  # noqa: N802
    return _EncodeAny("buttonset", lambda s: {}[s], DefButtonSetDict, buttonset)  # type: ignore[return-value]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..core.eudfunc.eudf import _EUDPredefineParam
from . import readtable

cpcache = c.curpl.GetCPCache()
_copydwn = c.Forward()
_check_copydwn = c.Memory(_copydwn, c.Exactly, 0)
_copydwn << _check_copydwn + 0
_read_end = readtable.read_end_common


@_EUDPredefineParam((ut.EPD(_read_end) + 86, ut.EPD(0x6509B0), ut.EPD(_copydwn)))
@c.EUDFunc
def f_repmovsd_epd(dstepdp, srcepdp, copydwn):
    dwread_cp_start = readtable._get(0xFFFFFFFF, 0)
    loopstart, contpoint, loopend, funcend = (c.Forward() for _ in range(4))
    c.RawTrigger(
        actions=[
            c.SetNextPtr(loopstart, dwread_cp_start),
            c.SetNextPtr(_read_end, contpoint),
            cpcache.SetDest(ut.EPD(0x6509B0)),
            c.SetNextPtr(cpcache.GetVTable(), funcend),
        ]
    )
    loopstart << c.RawTrigger(
        nextptr=dwread_cp_start,
        conditions=_check_copydwn,
        actions=c.SetNextPtr(loopstart, loopend),
    )
    contpoint << c.RawTrigger(
        nextptr=loopstart,
        actions=[
            c.SetMemory(_read_end + 344, c.Add, 1),
            c.SetMemory(0x6509B0, c.Add, 1),
            c.SetMemory(_copydwn, c.Subtract, 1),
        ],
    )

    loopend << cpcache.GetVTable()
    funcend << c.NextTrigger()


@c.EUDFunc
def _memcpy(dst, src, copylen):
    from .rwcommon import br1, bw1

    br1.seekoffset(src)
    bw1.seekoffset(dst)

    if cs.EUDWhile()(copylen >= 1):
        b = br1.readbyte()
        bw1.writebyte(b)
        copylen -= 1
    cs.EUDEndWhile()


def f_memcpy(dst, src, copylen, **kwargs):
    if (
        isinstance(dst, int)
        and isinstance(src, int)
        and isinstance(copylen, int)
        and dst % 4 == src % 4 == copylen % 4 == 0
    ):
        return f_repmovsd_epd(ut.EPD(dst), ut.EPD(src), copylen // 4, **kwargs)
    return _memcpy(dst, src, copylen, **kwargs)


@c.EUDFunc
def f_memcmp(buf1, buf2, count):
    from .rwcommon import br1, br2

    br1.seekoffset(buf1)
    br2.seekoffset(buf2)

    if cs.EUDWhile()(count >= 1):
        count -= 1
        ch1 = br1.readbyte()
        ch2 = br2.readbyte()
        cs.EUDContinueIf(ch1 == ch2)
        c.EUDReturn(ch1 - ch2)
    cs.EUDEndWhile()

    c.EUDReturn(0)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.
from ...bindings._rust import allocator as al

RlocInt_C = al.RlocInt_C
RlocInt = al.RlocInt
toRlocInt = al.toRlocInt  # noqa: N816



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .eudlv import EUDLightBool, EUDLightVariable
from .eudv import (
    EP_SetRValueStrictMode,
    EUDCreateVariables,
    EUDVariable,
    IsEUDVariable,
    NonSeqCompute,
    SeqCompute,
    SetVariables,
    VProc,
)
from .eudxv import EUDXVariable

__all__ = [
    "EUDLightBool",
    "EUDLightVariable",
    "EP_SetRValueStrictMode",
    "EUDCreateVariables",
    "EUDVariable",
    "IsEUDVariable",
    "NonSeqCompute",
    "SeqCompute",
    "SetVariables",
    "VProc",
    "EUDXVariable",
]



# Copyright 2024 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from abc import abstractmethod
from typing import Final, Literal, Self, overload

from ... import core as c
from ... import ctrlstru as cs
from ...core import EUDVariable
from ...localize import _
from ...utils import EPError, ExprProxy, ep_assert, unProxy
from .epdoffsetmap import EPDOffsetMap
from .member import BaseMember
from .memberkind import BaseKind, ByteKind, DwordKind, WordKind


class EnumMember(BaseMember, ExprProxy):
    __slots__ = (
        "_instance",
        "layout",
        "offset",
        "stride",
        "__objclass__",
        "__name__",
    )

    @property
    @abstractmethod
    def kind(self) -> type[BaseKind]:
        ...

    def __init__(
        self,
        layout: Literal["struct", "array"],
        offset: int,
        *,
        stride: int | None = None,
    ) -> None:
        self._instance: EPDOffsetMap | None = None  # FIXME
        super().__init__(layout, offset, stride=stride)
        super(BaseMember, self).__init__(None)

    def _get_epd(self, instance=None):
        if instance is None:
            instance = self._instance
        return super()._get_epd(instance)

    def __get__(
        self, instance: EPDOffsetMap | None, owner: type[EPDOffsetMap]
    ) -> Self:
        if instance is None:
            return self
        elif isinstance(instance, EPDOffsetMap):
            self._instance = instance
            # FIXME: fix reading value on every usages
            # example)
            # const movementFlags = unit.movementFlags;
            # arr[0] = movementFlags;
            # arr[1] = movementFlags;
            # unit.movementFlags += 1;
            # arr[2] = movementFlags;  // <- always read new value
            return self
        raise AttributeError

    def __set__(self, instance: EPDOffsetMap, value) -> None:
        self._instance = instance
        super().__set__(instance, value)

    # FIXME: Overwriting ExprProxy.getValue is kinda hacky...
    def getValue(self):  # noqa: N802
        epd, subp = self._get_epd()
        return self.kind.cast(self.kind.read_epd(epd, subp))

    def __lshift__(self, k):
        return self.getValue() << k

    def __rlshift__(self, k):
        return k << self.getValue()

    def __rshift__(self, k):
        return self.getValue() >> k

    def __rrshift__(self, k):
        return k >> self.getValue()

    def __add__(self, k):
        return self.getValue() + k

    def __radd__(self, k):
        return k + self.getValue()

    def __sub__(self, k):
        return self.getValue() - k

    def __rsub__(self, k):
        return k - self.getValue()

    def __mul__(self, k):
        return self.getValue() * k

    def __rmul__(self, k):
        return k * self.getValue()

    def __floordiv__(self, k):
        return self.getValue() // k

    def __rfloordiv__(self, k):
        return k // self.getValue()

    def __divmod__(self, k):
        return divmod(self.getValue(), k)

    def __rdivmod__(self, k):
        return divmod(k, self.getValue())

    def __mod__(self, k):
        return self.getValue() % k

    def __rmod__(self, k):
        return k % self.getValue()

    def __and__(self, k):
        return self.getValue() & k

    def __rand__(self, k):
        return k & self.getValue()

    def __or__(self, k):
        return self.getValue() | k

    def __ror__(self, k):
        return k | self.getValue()

    def __xor__(self, k):
        return self.getValue() ^ k

    def __rxor__(self, k):
        return k ^ self.getValue()

    def __neg__(self):
        return -self.getValue()

    def __invert__(self):
        return ~self.getValue()

    def __bool__(self):
        return bool(self.getValue())

    # Proxy comparison operator

    def __eq__(self, k):
        return self.getValue() == k

    def __ne__(self, k):
        return self.getValue() != k

    def __le__(self, k):
        return self.getValue() <= k

    def __lt__(self, k):
        return self.getValue() < k

    def __ge__(self, k):
        return self.getValue() >= k

    def __gt__(self, k):
        return self.getValue() > k


class ByteEnumMember(EnumMember):
    __slots__ = ()

    @property
    def kind(self):
        return ByteKind


class WordEnumMember(EnumMember):
    __slots__ = ()

    @property
    def kind(self):
        return WordKind


class DwordEnumMember(EnumMember):
    __slots__ = ()

    @property
    def kind(self):
        return DwordKind


class Flag:
    __slots__ = ("mask", "__objclass__", "__name__")

    mask: Final[int]

    def __init__(self, mask: int) -> None:
        ep_assert(0 < mask <= 0xFFFFFFFF, f"Invalid bitmask: {mask}")
        self.mask = mask

    @overload
    def __get__(self, instance: None, owner: type[EnumMember]) -> Self:
        ...

    @overload
    def __get__(self, instance: EnumMember, owner: type[EnumMember]) -> EUDVariable:
        ...

    def __get__(self, instance, owner=None) -> EUDVariable | Self:
        if instance is None:
            return self
        if instance.layout == "struct":
            from ...trigger import Trigger

            epd, subp = instance._get_epd()
            mask = self.mask << (8 * subp)
            ret = EUDVariable()
            ret << True
            Trigger(
                c.MemoryXEPD(epd, c.AtMost, mask - 1, mask), ret.SetNumber(False)
            )
            return ret
        elif instance.layout == "array":
            # FIXME: replace to efficent implementation
            flags = instance.getValue()
            ret = EUDVariable()
            ret << True
            c.RawTrigger(
                conditions=flags.AtMostX(self.mask - 1, self.mask),
                actions=ret.SetNumber(False),
            )
            return ret

        raise AttributeError

    def __set__(self, instance: EnumMember, value) -> None:
        from ...memio import f_maskwrite_epd

        # FIXME: replace to efficent implementation
        # while subp is always int for StructMember,
        # subp can be EUDVariable or int for ArrayMember
        epd, subp = instance._get_epd()
        mask = self.mask << (8 * subp)

        unproxied = unProxy(value)
        if isinstance(unproxied, EUDVariable):
            f_maskwrite_epd(epd, ~0, mask)
            if cs.EUDIfNot()(value):
                f_maskwrite_epd(epd, 0, mask)
            cs.EUDEndIf()
        elif unproxied in [0, False]:
            f_maskwrite_epd(epd, 0, mask)
        elif unproxied in [1, True]:
            f_maskwrite_epd(epd, ~0, mask)
        else:
            raise EPError(
                _("Can't assign {} to {}").format(
                    value, f"{self.__objclass__}.{self.__name__}"
                )
            )

    def __set_name__(self, owner: type[EnumMember], name: str) -> None:
        self.__objclass__ = owner
        self.__name__ = name


class MovementFlags(ByteEnumMember):
    """Flags that enable/disable certain movement."""

    __slots__ = ()
    OrderedAtLeastOnce = Flag(0x01)
    Accelerating = Flag(0x02)
    Braking = Flag(0x04)
    StartingAttack = Flag(0x08)
    Moving = Flag(0x10)
    Lifted = Flag(0x20)
    BrakeOnPathStep = Flag(0x40)
    "unit decelerates when reaching the end of current path segment"
    AlwaysZero = Flag(0x80)
    HoverUnit = Flag(0xC1)



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import itertools

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..eudlib.utilf.userpl import IsUserCP, f_getuserplayerid
from ..memio import (
    f_dwepdread_epd,
    f_dwwrite,
    f_epdread_epd,
    f_repmovsd_epd,
    f_setcurpl,
    f_wread_epd,
)
from ..scdata import CurrentPlayer
from .cpprint import f_cpstr_print
from .eudprint import f_dbstr_print
from .fmtprint import f_sprintf

_STAT_TEXT_POINTER = ut.EPD(0x6D5A30)  # TODO: hardcode 0x19184660 ?


@c.EUDTypedFunc([c.StatText])
def GetTBLAddr(tbl_id):  # noqa: N802
    add_tblptr, add_tblepd = c.Forward(), c.Forward()
    check_pointer = c.MemoryEPD(_STAT_TEXT_POINTER, c.Exactly, 0)
    if cs.EUDIfNot()(check_pointer):
        from ..core.variable.evcommon import _ev

        f_dwepdread_epd(_STAT_TEXT_POINTER, ret=[_ev[4], ut.EPD(add_tblepd) + 5])
        c.VProc(_ev[4], _ev[4].SetDest(ut.EPD(check_pointer) + 2))
        c.VProc(_ev[4], _ev[4].SetDest(ut.EPD(add_tblptr) + 5))
    cs.EUDEndIf()
    r, m = c.f_div(tbl_id, 2)
    c.RawTrigger(conditions=m.Exactly(1), actions=m.SetNumber(2))
    c.RawTrigger(actions=add_tblepd << r.AddNumber(0))
    ret = f_wread_epd(r, m)
    c.RawTrigger(actions=add_tblptr << ret.AddNumber(0))
    c.EUDReturn(ret)


def f_settbl(tbl_id, offset, *args, encoding=None):
    args = ut.FlattenList(args)
    dst = GetTBLAddr(tbl_id)
    dst += offset

    if encoding is None:
        encoding = c.StatText._encoding

    if encoding.casefold() == "UTF-8".casefold():
        if isinstance(args[-1], str):
            if args[-1][-1] != "\u2009":
                args[-1] = args[-1] + "\u2009"
        else:
            args.append("\u2009")

    return f_dbstr_print(dst, *args, encoding=encoding)


def f_settbl2(tbl_id, offset, *args, encoding=None):
    dst = GetTBLAddr(tbl_id)
    dst += offset

    if encoding is None:
        encoding = c.StatText._encoding

    return f_dbstr_print(dst, *args, EOS=False, encoding=encoding)


def f_settblf(tbl_id, offset, format_string, *args, encoding=None):
    dst = GetTBLAddr(tbl_id)
    dst += offset

    if encoding is None:
        encoding = c.StatText._encoding

    if encoding.casefold() == "UTF-8".casefold() and format_string[-1] != "\u2009":
        format_string += "\u2009"

    return f_sprintf(dst, format_string, *args, encoding=encoding)


def f_settblf2(tbl_id, offset, format_string, *args, encoding=None):
    dst = GetTBLAddr(tbl_id)
    dst += offset

    if encoding is None:
        encoding = c.StatText._encoding

    return f_sprintf(dst, format_string, *args, EOS=False, encoding=encoding)


_eprintln2_template = None
_eprintln2_print = c.Forward()
_eprintln2_eos = c.Forward()
_eprintln2_end = c.Forward()


def f_eprintln2(*args) -> None:
    global _eprintln2_template
    if _eprintln2_template is None:
        _eprintln2_template = c.Forward()

        c.PushTriggerScope()
        _eprintln2_template << c.NextTrigger()

        if cs.EUDExecuteOnce()():
            # [871] Unit's waypoint list is full.
            ptr = GetTBLAddr(871)
            f_dwwrite(ptr, ut.b2i4(b"\r\r\r\r"))
            epd = ut.EPD(ptr) + 1
        cs.EUDEndExecuteOnce()

        if cs.EUDIf()(IsUserCP()):
            f_setcurpl(epd)
            _eprintln2_print << c.RawTrigger(nextptr=0)
            _eprintln2_eos << c.RawTrigger(
                actions=c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0)
            )
            f_setcurpl(f_getuserplayerid())
        cs.EUDEndIf()
        _eprintln2_end << c.RawTrigger(nextptr=0)
        c.PopTriggerScope()

    _print, _next = c.Forward(), c.Forward()
    c.RawTrigger(
        nextptr=_eprintln2_template,
        actions=[
            c.SetNextPtr(_eprintln2_print, _print),
            c.SetNextPtr(_eprintln2_end, _next),
        ],
    )
    _print << c.NextTrigger()
    f_cpstr_print(*args, EOS=False, encoding="UTF-8")
    c.SetNextTrigger(_eprintln2_eos)
    _next << c.NextTrigger()


def _f_initstattext() -> None:
    if not c.StatText._data:
        return

    stattext_db = c.Db(c.StatText._data)
    inputdword_n = (len(c.StatText._data) + 3) // 4

    addr_epd = f_epdread_epd(ut.EPD(0x6D5A30))
    f_repmovsd_epd(addr_epd, ut.EPD(stattext_db), inputdword_n)


def _AddStatText(tbl: bytes) -> None:  # noqa: N802
    tbl_count = ut.b2i2(tbl)
    new_tbl = [ut.i2b2(tbl_count)]
    tbl_offset = 2 * (tbl_count + 1)
    if tbl_count % 2 == 0:
        tbl_offset += 2
    tbl_contents = []
    utf8_decodable_count = tbl_count
    cp949_decodable_count = tbl_count

    for i in range(tbl_count):
        k = 2 * (i + 1)
        tbl_start = ut.b2i2(tbl, k)
        tbl_end = ut.b2i2(tbl, k + 2)
        if i == tbl_count - 1:
            tbl_end = len(tbl)
        tbl_content = tbl[tbl_start:tbl_end]

        try:
            tbl_content.decode("utf-8")
        except UnicodeDecodeError:
            utf8_decodable_count -= 1
        try:
            tbl_content.decode("cp949")
        except UnicodeDecodeError:
            cp949_decodable_count -= 1

        assert tbl_offset % 4 == 0
        new_tbl.append(ut.i2b2(tbl_offset))
        new_tbl_len = -(-(len(tbl_content) + 3) // 4) * 4
        tbl_offset += new_tbl_len
        tbl_contents.append(tbl_content)
        tbl_contents.append(b"\0" * (new_tbl_len - len(tbl_content)))

    if tbl_count % 2 == 0:
        new_tbl.append(b"\0\0")

    if utf8_decodable_count <= cp949_decodable_count:
        c.StatText._encoding = "CP949"
    else:
        c.StatText._encoding = "UTF-8"

    c.StatText._data = b"".join(itertools.chain(new_tbl, tbl_contents))



import time

import helper
import profile_tool

# ruff: noqa: I001
# fmt: off

start_time = time.time()
DoCoverageTest = False

if DoCoverageTest:
    import coverage

    cov = coverage.Coverage()
    cov.start()

# helper._use_ptr_array()
helper.EP_SetRValueStrictMode(True)

from unittests import (
    testblockstru,
    testcurpl,
    testpatch,
    testarray,
    testptrigger,
    testoperator,
    testptrjump,
    testfptr,
    testprint,
    testswitch,
    testvartrg,
    testxvartrg,
    testmultiret,
    testvarray,
    testpvariable,
    teststruct,
    teststrfunc,
    testmath,
    testbinsearch,
    testbitwise,
    teststack,
    testdwmemio,
    testcpmemio,
    testptrmemio,
    testtypedf,
    testpool,
    testpoolfptr,
    test_lvalue,
    test_sq_from_1var,
    test_eps,
    testshortcircuit,
    testlistloopcompiles,
    testmapdatahelper,
    test_trace,
    test_pexists,
    test_dbstring,
    test_ctypes,
    test_dict_typo,

    test_arithmetics,
    test_fmtprint,
    test_locf,
    test_once,
    test_parse,
    test_scdata,
    test_unitgroup,
    testcondition,
)
# fmt: on

helper.CompressPayload(True)
helper.ShufflePayload(False)


def f():
    helper.test_runall("unittest")


from eudplib.string.tblprint import _AddStatText


with open("unittests/custom_txt.tbl", "rb") as tblfile:
    _AddStatText(tblfile.read())


# profile_tool.profile(f, "profile.json")
f()
print("--- %s seconds ---" % (time.time() - start_time))

if DoCoverageTest:
    cov.stop()
    cov.html_report(include=["C:\\eudplib\\*"])



from helper import *


@TestInstance
def test_sq_from_1var():
    a, b, c, d, e = EUDCreateVariables(5)
    e << 1234
    SeqCompute(
        [
            (a, SetTo, e),
            (b, SetTo, 1),
            (c, SetTo, 3),
            (d, SetTo, e),
            (e, SetTo, c),
            (d, SetTo, e),
        ]
    )
    test_equality("SeqCompute from 1 variable", [a, b, c, d, e], [1234, 1, 3, 3, 3])



from typing import Literal, TypeAlias

DefPortraitDict: dict[str, int] = {
    "Marine": 0,
    "Ghost": 1,
    "Firebat": 2,
    "Vulture": 3,
    "Spider Mine": 4,
    "Goliath": 5,
    "Siege Tank": 6,
    "SCV": 7,
    "Wraith": 8,
    "Science Vessel": 9,
    "Dropship": 10,
    "Battlecruiser": 11,
    "Sarah Kerrigan": 12,
    "Jim Raynor": 13,
    "Edmund Duke": 14,
    "Civilian": 15,
    "Arcturus Mengsk": 16,
    "Terran Advisor": 17,
    "Larva": 18,
    "Egg": 19,
    "Zergling": 20,
    "Hydralisk": 21,
    "Ultralisk": 22,
    "Broodling": 23,
    "Drone": 24,
    "Overlord": 25,
    "Mutalisk": 26,
    "Guardian": 27,
    "Queen": 28,
    "Defiler": 29,
    "Scourge": 30,
    "Cocoon": 31,
    "Mature Chrysalis": 32,
    "Infested Terran": 33,
    "Zasz": 34,
    "Daggoth": 35,
    "Infested Kerrigan": 36,
    "Hunter Killer": 37,
    "Overmind": 38,
    "Probe": 39,
    "Zealot": 40,
    "Dragoon": 41,
    "High Templar": 42,
    "Archon": 43,
    "Shuttle": 44,
    "Scout": 45,
    "Arbiter": 46,
    "Carrier": 47,
    "Interceptor": 48,
    "Dark Templar": 49,
    "Fenix (Zealot)": 50,
    "Fenix (Dragoon)": 51,
    "Zeratul": 52,
    "Tassadar": 53,
    "Gantrithor": 54,
    "Observer": 55,
    "Reaver": 56,
    "Scarab": 57,
    "Khaydarin Crystal Formation": 58,
    "Aldaris": 59,
    "Protoss Advisor": 60,
    "Merc Biker": 61,
    "Rhynadon": 62,
    "Bengalaas": 63,
    "Ragnasaur": 64,
    "Cargo Ship": 65,
    "Merc Gunship": 66,
    "Raider": 67,
    "Grom Biker": 68,
    "Space Critter": 69,
    "Sally": 70,
    "Greedo": 71,
    "Boskk": 72,
    "Peter": 73,
    "Independent Advisor": 74,
    "Terran Gas Tank": 75,
    "Protoss Gas Orb": 76,
    "Zerg Gas Sac": 77,
    "Mineral Cluster": 78,
    "Data Disc": 79,
    "Psi Emitter": 80,
    "Khaydarin Crystal": 81,
    # Flag color follows player color, not fixed.
    # "Flag (Red)": 82,
    # "Flag (Blue)": 83,
    # "Flag (Teal)": 84,
    # "Flag (Purple)": 85,
    # "Flag (Orange)": 86,
    # "Flag (Brown)": 87,
    # "Flag (White)": 88,
    # "Flag (Yellow)": 89,
    "Flag P1": 82,
    "Flag P2": 83,
    "Flag P3": 84,
    "Flag P4": 85,
    "Flag P5": 86,
    "Flag P6": 87,
    "Flag P7": 88,
    "Flag P8": 89,
    "Medic": 90,
    "Valkyrie": 91,
    "Dugalle": 92,
    "Stukov": 93,
    "Duran": 94,
    "Artanis": 95,
    "Raszagal": 96,
    "Devourer": 97,
    "Lurker": 98,
    "Dark Archon": 99,
    "Corsair": 100,
    "Scantid": 101,
    "Kakaru": 102,
    "Ursadon": 103,
    "Uraj Crystal": 104,
    "Khalis Crystal": 105,
    # "Flag (Green)": 106,
    # "Flag (Bright Yellow)": 107,
    # "Flag (Tan)": 108,
    # "Flag (Blue)": 109,
    "Flag P9": 106,
    "Flag P10": 107,
    "Flag P11": 108,
    "Flag P12": 109,
    "Marine - Talking": 110,
    "Ghost - Talking": 111,
    "Firebat - Talking": 112,
    "Vulture - Talking": 113,
    "Spider Mine - Talking": 114,
    "Goliath - Talking": 115,
    "Siege Tank - Talking": 116,
    "SCV - Talking": 117,
    "Wraith - Talking": 118,
    "Science Vessel - Talking": 119,
    "Dropship - Talking": 120,
    "Battlecruiser - Talking": 121,
    "Sarah Kerrigan - Talking": 122,
    "Jim Raynor  - Talking": 123,
    "Edmund Duke - Talking": 124,
    "Civilian - Talking": 125,
    "Arcturus Mengsk - Talking": 126,
    "Terran Advisor - Talking": 127,
    "Larva - Talking": 128,
    "Egg - Talking": 129,
    "Zergling - Talking": 130,
    "Hydralisk - Talking": 131,
    "Ultralisk - Talking": 132,
    "Broodling - Talking": 133,
    "Drone - Talking": 134,
    "Overlord - Talking": 135,
    "Mutalisk - Talking": 136,
    "Guardian - Talking": 137,
    "Queen - Talking": 138,
    "Defiler - Talking": 139,
    "Scourge - Talking": 140,
    "Cocoon - Talking": 141,
    "Mature Chrysalis - Talking": 142,
    "Infested Terran - Talking": 143,
    "Zasz - Talking": 144,
    "Daggoth - Talking": 145,
    "Infested Kerrigan - Talking": 146,
    "Hunter Killer - Talking": 147,
    "Overmind - Talking": 148,
    "Probe - Talking": 149,
    "Zealot - Talking": 150,
    "Dragoon - Talking": 151,
    "High Templar - Talking": 152,
    "Archon - Talking": 153,
    "Shuttle - Talking": 154,
    "Scout - Talking": 155,
    "Arbiter - Talking": 156,
    "Carrier - Talking": 157,
    "Interceptor - Talking": 158,
    "Dark Templar - Talking": 159,
    "Fenix (Zealot) - Talking": 160,
    "Fenix (Dragoon) - Talking": 161,
    "Zeratul - Talking": 162,
    "Tassadar - Talking": 163,
    "Gantrithor - Talking": 164,
    "Observer - Talking": 165,
    "Reaver - Talking": 166,
    "Scarab - Talking": 167,
    "Khaydarin Crystal Formation - Talking": 168,
    "Aldaris - Talking": 169,
    "Protoss Advisor - Talking": 170,
    "Merc Biker - Talking": 171,
    "Rhynadon - Talking": 172,
    "Bengalaas - Talking": 173,
    "Ragnasaur - Talking": 174,
    "Cargo Ship - Talking": 175,
    "Merc Gunship - Talking": 176,
    "Raider - Talking": 177,
    "Grom Biker - Talking": 178,
    "Space Critter - Talking": 179,
    "Sally - Talking": 180,
    "Greedo - Talking": 181,
    "Boskk - Talking": 182,
    "Peter - Talking": 183,
    "Independent Advisor - Talking": 184,
    "Terran Gas Tank - Talking": 185,
    "Protoss Gas Orb - Talking": 186,
    "Zerg Gas Sac - Talking": 187,
    "Mineral Cluster - Talking": 188,
    "Data Disc - Talking": 189,
    "Psi Emitter - Talking": 190,
    "Khaydarin Crystal - Talking": 191,
    # "Flag (Red) - Talking": 192,
    # "Flag (Blue) - Talking": 193,
    # "Flag (Teal) - Talking": 194,
    # "Flag (Purple) - Talking": 195,
    # "Flag (Orange) - Talking": 196,
    # "Flag (Brown) - Talking": 197,
    # "Flag (White) - Talking": 198,
    # "Flag (Yellow) - Talking": 199,
    "Flag P1 - Talking": 192,
    "Flag P2 - Talking": 193,
    "Flag P3 - Talking": 194,
    "Flag P4 - Talking": 195,
    "Flag P5 - Talking": 196,
    "Flag P6 - Talking": 197,
    "Flag P7 - Talking": 198,
    "Flag P8 - Talking": 199,
    "Medic - Talking": 200,
    "Valkyrie - Talking": 201,
    "Dugalle - Talking": 202,
    "Stukov - Talking": 203,
    "Duran - Talking": 204,
    "Artanis - Talking": 205,
    "Raszagal - Talking": 206,
    "Devourer - Talking": 207,
    "Lurker - Talking": 208,
    "Dark Archon - Talking": 209,
    "Corsair - Talking": 210,
    "Scantid - Talking": 211,
    "Kakaru - Talking": 212,
    "Ursadon - Talking": 213,
    "Uraj Crystal - Talking": 214,
    "Khalis Crystal - Talking": 215,
    # "Flag (Green) - Talking": 216,
    # "Flag (Bright Yellow) - Talking": 217,
    # "Flag (Tan) - Talking": 218,
    # "Flag (Blue) - Talkin": 219,
    "Flag P9 - Talking": 216,
    "Flag P10 - Talking": 217,
    "Flag P11 - Talking": 218,
    "Flag P12 - Talking": 219,
    "None": 65535,
}

DefaultPortrait: TypeAlias = Literal[
    "Marine",
    "Ghost",
    "Firebat",
    "Vulture",
    "Spider Mine",
    "Goliath",
    "Siege Tank",
    "SCV",
    "Wraith",
    "Science Vessel",
    "Dropship",
    "Battlecruiser",
    "Sarah Kerrigan",
    "Jim Raynor",
    "Edmund Duke",
    "Civilian",
    "Arcturus Mengsk",
    "Terran Advisor",
    "Larva",
    "Egg",
    "Zergling",
    "Hydralisk",
    "Ultralisk",
    "Broodling",
    "Drone",
    "Overlord",
    "Mutalisk",
    "Guardian",
    "Queen",
    "Defiler",
    "Scourge",
    "Cocoon",
    "Mature Chrysalis",
    "Infested Terran",
    "Zasz",
    "Daggoth",
    "Infested Kerrigan",
    "Hunter Killer",
    "Overmind",
    "Probe",
    "Zealot",
    "Dragoon",
    "High Templar",
    "Archon",
    "Shuttle",
    "Scout",
    "Arbiter",
    "Carrier",
    "Interceptor",
    "Dark Templar",
    "Fenix (Zealot)",
    "Fenix (Dragoon)",
    "Zeratul",
    "Tassadar",
    "Gantrithor",
    "Observer",
    "Reaver",
    "Scarab",
    "Khaydarin Crystal Formation",
    "Aldaris",
    "Protoss Advisor",
    "Merc Biker",
    "Rhynadon",
    "Bengalaas",
    "Ragnasaur",
    "Cargo Ship",
    "Merc Gunship",
    "Raider",
    "Grom Biker",
    "Space Critter",
    "Sally",
    "Greedo",
    "Boskk",
    "Peter",
    "Independent Advisor",
    "Terran Gas Tank",
    "Protoss Gas Orb",
    "Zerg Gas Sac",
    "Mineral Cluster",
    "Data Disc",
    "Psi Emitter",
    "Khaydarin Crystal",
    "Flag P1",
    "Flag P2",
    "Flag P3",
    "Flag P4",
    "Flag P5",
    "Flag P6",
    "Flag P7",
    "Flag P8",
    "Medic",
    "Valkyrie",
    "Dugalle",
    "Stukov",
    "Duran",
    "Artanis",
    "Raszagal",
    "Devourer",
    "Lurker",
    "Dark Archon",
    "Corsair",
    "Scantid",
    "Kakaru",
    "Ursadon",
    "Uraj Crystal",
    "Khalis Crystal",
    "Flag P9",
    "Flag P10",
    "Flag P11",
    "Flag P12",
    "Marine - Talking",
    "Ghost - Talking",
    "Firebat - Talking",
    "Vulture - Talking",
    "Spider Mine - Talking",
    "Goliath - Talking",
    "Siege Tank - Talking",
    "SCV - Talking",
    "Wraith - Talking",
    "Science Vessel - Talking",
    "Dropship - Talking",
    "Battlecruiser - Talking",
    "Sarah Kerrigan - Talking",
    "Jim Raynor  - Talking",
    "Edmund Duke - Talking",
    "Civilian - Talking",
    "Arcturus Mengsk - Talking",
    "Terran Advisor - Talking",
    "Larva - Talking",
    "Egg - Talking",
    "Zergling - Talking",
    "Hydralisk - Talking",
    "Ultralisk - Talking",
    "Broodling - Talking",
    "Drone - Talking",
    "Overlord - Talking",
    "Mutalisk - Talking",
    "Guardian - Talking",
    "Queen - Talking",
    "Defiler - Talking",
    "Scourge - Talking",
    "Cocoon - Talking",
    "Mature Chrysalis - Talking",
    "Infested Terran - Talking",
    "Zasz - Talking",
    "Daggoth - Talking",
    "Infested Kerrigan - Talking",
    "Hunter Killer - Talking",
    "Overmind - Talking",
    "Probe - Talking",
    "Zealot - Talking",
    "Dragoon - Talking",
    "High Templar - Talking",
    "Archon - Talking",
    "Shuttle - Talking",
    "Scout - Talking",
    "Arbiter - Talking",
    "Carrier - Talking",
    "Interceptor - Talking",
    "Dark Templar - Talking",
    "Fenix (Zealot) - Talking",
    "Fenix (Dragoon) - Talking",
    "Zeratul - Talking",
    "Tassadar - Talking",
    "Gantrithor - Talking",
    "Observer - Talking",
    "Reaver - Talking",
    "Scarab - Talking",
    "Khaydarin Crystal Formation - Talking",
    "Aldaris - Talking",
    "Protoss Advisor - Talking",
    "Merc Biker - Talking",
    "Rhynadon - Talking",
    "Bengalaas - Talking",
    "Ragnasaur - Talking",
    "Cargo Ship - Talking",
    "Merc Gunship - Talking",
    "Raider - Talking",
    "Grom Biker - Talking",
    "Space Critter - Talking",
    "Sally - Talking",
    "Greedo - Talking",
    "Boskk - Talking",
    "Peter - Talking",
    "Independent Advisor - Talking",
    "Terran Gas Tank - Talking",
    "Protoss Gas Orb - Talking",
    "Zerg Gas Sac - Talking",
    "Mineral Cluster - Talking",
    "Data Disc - Talking",
    "Psi Emitter - Talking",
    "Khaydarin Crystal - Talking",
    "Flag P1 - Talking",
    "Flag P2 - Talking",
    "Flag P3 - Talking",
    "Flag P4 - Talking",
    "Flag P5 - Talking",
    "Flag P6 - Talking",
    "Flag P7 - Talking",
    "Flag P8 - Talking",
    "Medic - Talking",
    "Valkyrie - Talking",
    "Dugalle - Talking",
    "Stukov - Talking",
    "Duran - Talking",
    "Artanis - Talking",
    "Raszagal - Talking",
    "Devourer - Talking",
    "Lurker - Talking",
    "Dark Archon - Talking",
    "Corsair - Talking",
    "Scantid - Talking",
    "Kakaru - Talking",
    "Ursadon - Talking",
    "Uraj Crystal - Talking",
    "Khalis Crystal - Talking",
    "Flag P9 - Talking",
    "Flag P10 - Talking",
    "Flag P11 - Talking",
    "Flag P12 - Talking",
    "None",
]



# Copyright 2024 by zzt (Defender).
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: N815
from typing import ClassVar

from .. import utils as ut
from ..core.rawtrigger.consttype import ConstType
from ..core.rawtrigger.strenc import EncodeFlingy
from ..core.rawtrigger.typehint import _Flingy
from ..localize import _
from .offsetmap import (
    ByteMember,
    DwordMember,
    EPDOffsetMap,
    MovementControlMember,
    SpriteMember,
    WordMember,
)


class Flingy(EPDOffsetMap, ConstType):
    __slots__ = ()
    sprite: ClassVar[SpriteMember] = SpriteMember("array", 0x6CA318)
    topSpeed: ClassVar[DwordMember] = DwordMember("array", 0x6C9EF8)
    acceleration: ClassVar[WordMember] = WordMember("array", 0x6C9C78)
    haltDistance: ClassVar[DwordMember] = DwordMember("array", 0x6C9930)
    turnSpeed: ClassVar[ByteMember] = ByteMember("array", 0x6C9E20)
    turnRadius = turnSpeed
    # unused: ClassVar[ByteMember] = ByteMember("array", 0x6CA240)
    movementControl: ClassVar[MovementControlMember] = MovementControlMember(
        "array", 0x6C9858
    )

    @ut.classproperty
    def range(self):
        return (0, 208, 1)

    @classmethod
    def cast(cls, _from: _Flingy):
        if isinstance(_from, ConstType) and not isinstance(_from, cls):
            raise ut.EPError(_('"{}" is not a {}').format(_from, cls.__name__))
        return super().cast(_from)

    def __init__(self, initval: _Flingy) -> None:
        super().__init__(EncodeFlingy(initval))



from helper import *


# Basic struct test
class TestStruct(EUDStruct):
    _fields_ = ["x", "y", "z"]

    @EUDMethod
    def setX(self, x):
        self.x = x

    @EUDTypedMethod([selftype])
    def add(self, other):
        self.x += other.x
        self.y += other.y
        self.z += other.z
        return self


@TestInstance
def test_struct():
    a = TestStruct()
    a.x = 5
    a.y = 7

    b = TestStruct.cast(a)
    c = TestStruct.cast(b)
    d = TestStruct.cast(c)

    b.z = 8
    c.x = 3

    test_assert("EUDStruct can be converted to a condition", a)

    test_equality(
        "EUDStruct indirect access test",
        [a.x, b.x, c.x, d.x, a.y, b.y, c.y, d.y, a.z, b.z, c.z, d.z],
        [3, 3, 3, 3, 7, 7, 7, 7, 8, 8, 8, 8],
    )

    d = a.copy()
    a.add(d)
    test_equality(
        "EUDTypedMethod test", [a.x, a.y, a.z, d.x, d.y, d.z], [6, 14, 16, 3, 7, 8]
    )

    e = EUDVariable()
    e << a
    TestStruct.cast(e).setX(1)
    test_equality("EUDStruct method via EUDVariable", a.x, 1)

    # https://github.com/phu54321/euddraft/issues/8
    # Accessing to unknown fields should trigger an error.
    with expect_eperror():
        a.q = 1


# Nested struct test


class Coord(EUDStruct):
    _fields_ = ["x", "y"]


class Triangle(EUDStruct):
    _fields_ = [("p", Coord * 5), "q"]


@TestInstance
def test_nested_struct():
    a = Triangle()
    a.p = (Coord * 5)()
    for i in range(5):
        a.p[i] = Coord()

    a.p[0].x = 1
    a.p[1].y = 2
    a.p[1].y += 2
    a.q = 3

    test_assert(
        "Nested EUDStruct test",
        [a.p[0].x == 1, a.p[0].y == 0, a.p[1].y == 4, a.q == 3],
    )

    b = a.copy()
    b.p[0].x = 5
    b.q = 2
    test_equality(
        "Shallow EUDStruct copy test",
        [a.p[0].x, b.p[0].x, b.p[0].y, b.p[1].y, a.q, b.q],
        [5, 5, 0, 4, 3, 2],
    )

    point3x2 = (Coord * 3 * 2)(
        [(Coord * 3)([Coord() for _ in range(3)]) for _ in range(2)]
    )

    with expect_eperror():
        point3x2[2]
    with expect_eperror():
        point3x2[0][3]

    point3x2[0][0].x = 1
    point3x2[0][0].y = 2
    point3x2[1][0].x = 3
    point3x2[0][1].y = 4
    point3x2[1][1].x = 5
    point3x2[1][1].y = 6

    test_equality(
        "Nested EUDStructArray test",
        [
            point3x2[0][0].x,
            point3x2[0][0].y,
            point3x2[0][1].x,
            point3x2[0][1].y,
            point3x2[1][0].x,
            point3x2[1][0].y,
            point3x2[1][1].x,
            point3x2[1][1].y,
        ],
        [1, 2, 0, 4, 3, 0, 5, 6],
    )


@TestInstance
def test_invalid_struct():
    with expect_eperror():
        a = TestStruct()
        a << 5
    with expect_error(TypeError):
        TestStruct()[0]
    with expect_error(TypeError):
        TestStruct()[0] = 0
    with expect_error(TypeError):
        for _ in TestStruct():
            pass


@TestInstance
def test_cast_0_to_struct():
    a = TestStruct.cast(0)
    test_equality("nullptr casting", a, 0)


# Test referencing itself on member decl
class ListClass(EUDStruct):
    _fields_ = [("prev", selftype), ("next", selftype)]


@TestInstance
def test_selftype_member():
    a = ListClass()
    b = ListClass()
    a.prev = a.next = b
    a.prev = b.next = a



from helper import *


@TestInstance
def test_locf():
    loc = EUDVariable(1)

    f_setloc(loc, 32, 64)
    f_addloc(loc, 16, 8)
    f_dilateloc(loc, 4, 2)

    L, T = f_getlocTL(loc)

    test_equality("f_getlocTL test", [L, T], [44, 70])
    test_equality(
        "Location functions test",
        [
            f_dwread_epd(EPD(0x58DC68)),
            f_dwread_epd(EPD(0x58DC6C)),
        ],
        [52, 74],
    )

    f_setloc(1, EUDVariable(51), EUDVariable(76))
    test_equality(
        "f_setloc test",
        [
            f_dwread_epd(EPD(0x58DC68)),
            f_dwread_epd(EPD(0x58DC6C)),
        ],
        [51, 76],
    )

    Pos = Db(i2b4(0x6FE05BD))
    f_setloc_epd(loc, EPD(Pos))
    L, T = f_getlocTL(loc)
    test_equality("f_setloc_epd test", [L, T], [0x5BD, 0x6FE])



from typing import Literal, TypeAlias

DefFlingyDict: dict[str, int] = {
    "Scourge": 0,
    "Broodling": 1,
    "Infested Terran": 2,
    "Guardian Cocoon": 3,
    "Defiler": 4,
    "Drone": 5,
    "Egg": 6,
    "Guardian": 7,
    "Hydralisk": 8,
    "Infested Kerrigan": 9,
    "Larva": 10,
    "Mutalisk": 11,
    "Overlord": 12,
    "Queen": 13,
    "Ultralisk": 14,
    "Zergling": 15,
    "Cerebrate": 16,
    "Infested Command Center": 17,
    "Spawning Pool": 18,
    "Mature Chrysalis": 19,
    "Evolution Chamber": 20,
    "Creep Colony": 21,
    "Hatchery": 22,
    "Hive": 23,
    "Lair": 24,
    "Sunken Colony": 25,
    "Greater Spire": 26,
    "Defiler Mound": 27,
    "Queen's Nest": 28,
    "Nydus Canal": 29,
    "Overmind With Shell": 30,
    "Overmind Without Shell": 31,
    "Ultralisk Cavern": 32,
    "Extractor": 33,
    "Hydralisk Den": 34,
    "Spire": 35,
    "Spore Colony": 36,
    "Arbiter": 37,
    "Archon Energy": 38,
    "Carrier": 39,
    "Dragoon": 40,
    "Interceptor": 41,
    "Probe": 42,
    "Scout": 43,
    "Shuttle": 44,
    "High Templar": 45,
    "Dark Templar (Hero)": 46,
    "Reaver": 47,
    "Scarab": 48,
    "Zealot": 49,
    "Observer": 50,
    "Templar Archives": 51,
    "Assimilator": 52,
    "Observatory": 53,
    "Citadel of Adun": 54,
    "Forge": 55,
    "Gateway": 56,
    "Cybernetics Core": 57,
    "Khaydarin Crystal Formation": 58,
    "Nexus": 59,
    "Photon Cannon": 60,
    "Arbiter Tribunal": 61,
    "Pylon": 62,
    "Robotics Facility": 63,
    "Shield Battery": 64,
    "Stargate": 65,
    "Stasis Cell/Prison": 66,
    "Robotics Support Bay": 67,
    "Protoss Temple": 68,
    "Fleet Beacon": 69,
    "Battlecruiser": 70,
    "Civilian": 71,
    "Dropship": 72,
    "Firebat": 73,
    "Ghost": 74,
    "Goliath Base": 75,
    "Goliath Turret": 76,
    "Sarah Kerrigan": 77,
    "Marine": 78,
    "Scanner Sweep": 79,
    "Wraith": 80,
    "SCV": 81,
    "Siege Tank (Tank) Base": 82,
    "Siege Tank (Tank) Turret": 83,
    "Siege Tank (Siege) Base": 84,
    "Siege Tank (Siege) Turret": 85,
    "Science Vessel (Base)": 86,
    "Science Vessel (Turret)": 87,
    "Vulture": 88,
    "Spider Mine": 89,
    "Academy": 90,
    "Barracks": 91,
    "Armory": 92,
    "Comsat Station": 93,
    "Command Center": 94,
    "Supply Depot": 95,
    "Control Tower": 96,
    "Factory": 97,
    "Covert Ops": 98,
    "Ion Cannon": 99,
    "Machine Shop": 100,
    "Missile Turret": 101,
    "Crashed Battlecruiser": 102,
    "Physics Lab": 103,
    "Bunker": 104,
    "Refinery": 105,
    "Immobile Barracks": 106,
    "Science Facility": 107,
    "Nuke Silo": 108,
    "Nuclear Missile": 109,
    "Starport": 110,
    "Engineering Bay": 111,
    "Terran Construction (Large)": 112,
    "Terran Construction (Small)": 113,
    "Ragnasaur": 114,
    "Rhynadon": 115,
    "Bengalaas": 116,
    "Vespene Geyser": 117,
    "Mineral Field Type1": 118,
    "Mineral Field Type2": 119,
    "Mineral Field Type3": 120,
    "Independent Starport": 121,
    "Zerg Beacon": 122,
    "Terran Beacon": 123,
    "Protoss Beacon": 124,
    "Dark Swarm": 125,
    "Flag": 126,
    "Young Chrysalis": 127,
    "Psi Emitter": 128,
    "Data Disc": 129,
    "Khaydarin Crystal": 130,
    "Mineral Chunk Type1": 131,
    "Mineral Chunk Type2": 132,
    "Protoss Gas Orb Type1": 133,
    "Protoss Gas Orb Type2": 134,
    "Zerg Gas Sac Type1": 135,
    "Zerg Gas Sac Type2": 136,
    "Terran Gas Tank Type1": 137,
    "Terran Gas Tank Type2": 138,
    "Map Revealer": 139,
    "Start Location": 140,
    "Fusion Cutter Hit": 141,
    "Gauss Rifle Hit": 142,
    "C-10 Canister Rifle Hit": 143,
    "Gemini Missiles": 144,
    "Fragmentation Grenade": 145,
    "Lockdown/LongBolt/Hellfire Missile": 146,
    "Unused Lockdown": 147,
    "ATS/ATA Laser Battery": 148,
    "Burst Lasers": 149,
    "Arclite Shock Cannon Hit": 150,
    "EMP Missile": 151,
    "Dual Photon Blasters Hit": 152,
    "Particle Beam Hit": 153,
    "Anti-Matter Missile": 154,
    "Pulse Cannon": 155,
    "Psionic Shockwave Hit": 156,
    "Psionic Storm": 157,
    "Yamato Gun": 158,
    "Phase Disruptor": 159,
    "STA/STS Photon Cannon Overlay": 160,
    "Sunken Colony Tentacle": 161,
    "Venom": 162,
    "Acid Spore": 163,
    "Unknown164": 164,
    "Glave Wurm": 165,
    "Seeker Spores": 166,
    "Queen Spell Carrier": 167,
    "Plague Cloud": 168,
    "Consume": 169,
    "Ensnare": 170,
    "Needle Spine Hit": 171,
    "White Circle (Invisible)": 172,
    "Left Upper Level Door": 173,
    "Right Upper Level Door": 174,
    "Substructure Left Door": 175,
    "Substructure Right Door": 176,
    "Substructure Opening Hole": 177,
    "Floor Gun Trap": 178,
    "Floor Missile Trap": 179,
    "Wall Missile Trap": 180,
    "Wall Missile Trap2": 181,
    "Wall Flame Trap": 182,
    "Wall Flame Trap2": 183,
    "Lurker Egg": 184,
    "Devourer": 185,
    "Lurker": 186,
    "Dark Archon Energy": 187,
    "Dark Templar (Unit)": 188,
    "Medic": 189,
    "Valkyrie": 190,
    "Corsair": 191,
    "Disruption Web": 192,
    "Overmind Cocoon": 193,
    "Psi Disrupter": 194,
    "Warp Gate": 195,
    "Power Generator": 196,
    "Xel'Naga Temple": 197,
    "Scantid": 198,
    "Kakaru": 199,
    "Ursadon": 200,
    "Optical Flare Grenade": 201,
    "Halo Rockets": 202,
    "Subterranean Spines": 203,
    "Corrosive Acid Shot": 204,
    "Corrosive Acid Hit": 205,
    "Neutron Flare": 206,
    "Uraj": 207,
    "Khalis": 208,
}

DefaultFlingy: TypeAlias = Literal[
    "Scourge",
    "Broodling",
    "Infested Terran",
    "Guardian Cocoon",
    "Defiler",
    "Drone",
    "Egg",
    "Guardian",
    "Hydralisk",
    "Infested Kerrigan",
    "Larva",
    "Mutalisk",
    "Overlord",
    "Queen",
    "Ultralisk",
    "Zergling",
    "Cerebrate",
    "Infested Command Center",
    "Spawning Pool",
    "Mature Chrysalis",
    "Evolution Chamber",
    "Creep Colony",
    "Hatchery",
    "Hive",
    "Lair",
    "Sunken Colony",
    "Greater Spire",
    "Defiler Mound",
    "Queen's Nest",
    "Nydus Canal",
    "Overmind With Shell",
    "Overmind Without Shell",
    "Ultralisk Cavern",
    "Extractor",
    "Hydralisk Den",
    "Spire",
    "Spore Colony",
    "Arbiter",
    "Archon Energy",
    "Carrier",
    "Dragoon",
    "Interceptor",
    "Probe",
    "Scout",
    "Shuttle",
    "High Templar",
    "Dark Templar (Hero)",
    "Reaver",
    "Scarab",
    "Zealot",
    "Observer",
    "Templar Archives",
    "Assimilator",
    "Observatory",
    "Citadel of Adun",
    "Forge",
    "Gateway",
    "Cybernetics Core",
    "Khaydarin Crystal Formation",
    "Nexus",
    "Photon Cannon",
    "Arbiter Tribunal",
    "Pylon",
    "Robotics Facility",
    "Shield Battery",
    "Stargate",
    "Stasis Cell/Prison",
    "Robotics Support Bay",
    "Protoss Temple",
    "Fleet Beacon",
    "Battlecruiser",
    "Civilian",
    "Dropship",
    "Firebat",
    "Ghost",
    "Goliath Base",
    "Goliath Turret",
    "Sarah Kerrigan",
    "Marine",
    "Scanner Sweep",
    "Wraith",
    "SCV",
    "Siege Tank (Tank) Base",
    "Siege Tank (Tank) Turret",
    "Siege Tank (Siege) Base",
    "Siege Tank (Siege) Turret",
    "Science Vessel (Base)",
    "Science Vessel (Turret)",
    "Vulture",
    "Spider Mine",
    "Academy",
    "Barracks",
    "Armory",
    "Comsat Station",
    "Command Center",
    "Supply Depot",
    "Control Tower",
    "Factory",
    "Covert Ops",
    "Ion Cannon",
    "Machine Shop",
    "Missile Turret",
    "Crashed Battlecruiser",
    "Physics Lab",
    "Bunker",
    "Refinery",
    "Immobile Barracks",
    "Science Facility",
    "Nuke Silo",
    "Nuclear Missile",
    "Starport",
    "Engineering Bay",
    "Terran Construction (Large)",
    "Terran Construction (Small)",
    "Ragnasaur",
    "Rhynadon",
    "Bengalaas",
    "Vespene Geyser",
    "Mineral Field Type1",
    "Mineral Field Type2",
    "Mineral Field Type3",
    "Independent Starport",
    "Zerg Beacon",
    "Terran Beacon",
    "Protoss Beacon",
    "Dark Swarm",
    "Flag",
    "Young Chrysalis",
    "Psi Emitter",
    "Data Disc",
    "Khaydarin Crystal",
    "Mineral Chunk Type1",
    "Mineral Chunk Type2",
    "Protoss Gas Orb Type1",
    "Protoss Gas Orb Type2",
    "Zerg Gas Sac Type1",
    "Zerg Gas Sac Type2",
    "Terran Gas Tank Type1",
    "Terran Gas Tank Type2",
    "Map Revealer",
    "Start Location",
    "Fusion Cutter Hit",
    "Gauss Rifle Hit",
    "C-10 Canister Rifle Hit",
    "Gemini Missiles",
    "Fragmentation Grenade",
    "Lockdown/LongBolt/Hellfire Missile",
    "Unused Lockdown",
    "ATS/ATA Laser Battery",
    "Burst Lasers",
    "Arclite Shock Cannon Hit",
    "EMP Missile",
    "Dual Photon Blasters Hit",
    "Particle Beam Hit",
    "Anti-Matter Missile",
    "Pulse Cannon",
    "Psionic Shockwave Hit",
    "Psionic Storm",
    "Yamato Gun",
    "Phase Disruptor",
    "STA/STS Photon Cannon Overlay",
    "Sunken Colony Tentacle",
    "Venom",
    "Acid Spore",
    "Unknown164",
    "Glave Wurm",
    "Seeker Spores",
    "Queen Spell Carrier",
    "Plague Cloud",
    "Consume",
    "Ensnare",
    "Needle Spine Hit",
    "White Circle (Invisible)",
    "Left Upper Level Door",
    "Right Upper Level Door",
    "Substructure Left Door",
    "Substructure Right Door",
    "Substructure Opening Hole",
    "Floor Gun Trap",
    "Floor Missile Trap",
    "Wall Missile Trap",
    "Wall Missile Trap2",
    "Wall Flame Trap",
    "Wall Flame Trap2",
    "Lurker Egg",
    "Devourer",
    "Lurker",
    "Dark Archon Energy",
    "Dark Templar (Unit)",
    "Medic",
    "Valkyrie",
    "Corsair",
    "Disruption Web",
    "Overmind Cocoon",
    "Psi Disrupter",
    "Warp Gate",
    "Power Generator",
    "Xel'Naga Temple",
    "Scantid",
    "Kakaru",
    "Ursadon",
    "Optical Flare Grenade",
    "Halo Rockets",
    "Subterranean Spines",
    "Corrosive Acid Shot",
    "Corrosive Acid Hit",
    "Neutron Flare",
    "Uraj",
    "Khalis",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import functools
import os.path
import random
import sys
from collections.abc import Iterable, Iterator, Sequence
from typing import Any, TypeVar, overload

from .. import core as c
from ..localize import _
from .eperror import EPError, ep_warn
from .exprproxy import unProxy

T = TypeVar("T")
_epd_on_epd = True


def _allow_epd_on_epd(b: bool) -> None:
    global _epd_on_epd
    _epd_on_epd = b


def EPD(p: Any, **kwargs) -> Any:  # noqa: N802
    if c.IsConstExpr(p):
        p = unProxy(p)
        epd: c.ConstExpr | int
        if isinstance(p, c.ConstExpr) and p._is_epd():
            if _epd_on_epd:
                ep_warn(_("EPD on EPD value of ConstExpr is no-op"))
                epd = p
            else:
                raise EPError(_("EPD on EPD value of ConstExpr is no-op"))
        else:
            epd = (p + (-0x58A364)) // 4
        if "ret" in kwargs:
            c.SeqCompute([(kwargs["ret"][0], c.SetTo, epd)])
            return kwargs["ret"][0]
        return epd

    if c.IsEUDVariable(p):
        if sys.getrefcount(p) <= 2:
            mask = 0b111
            actions = [
                p.AddNumber(-0x58A364),
                p.SetNumberX(0, mask >> 1),
            ]
            actions.extend(
                p.SubtractNumberX((mask >> 1) << t, mask << t) for t in range(30)
            )
            c.RawTrigger(actions=actions)
            return p
        if not hasattr(EPD, "_eudf"):
            c.PushTriggerScope()
            setter = c.SetDeaths(0, c.SetTo, 0, 0)
            vaddr = setter + 20
            mask = 0b111
            actions = [
                c.SetMemory(vaddr, c.Add, -0x58A364),
                c.SetMemoryX(vaddr, c.SetTo, 0, mask >> 1),
            ]
            actions.extend(
                c.SetMemoryX(vaddr, c.Subtract, (mask >> 1) << t, mask << t)
                for t in range(30)
            )
            actions.append(setter)
            ftrg = c.RawTrigger(nextptr=0, actions=actions)
            c.PopTriggerScope()

            setattr(EPD, "_eudf", (ftrg, setter))

        ret = kwargs["ret"][0] if "ret" in kwargs else c.EUDVariable()
        ftrg, setter = getattr(EPD, "_eudf")
        nexttrg = c.Forward()
        dst = EPD(ret.getValueAddr()) if c.IsEUDVariable(ret) else ret
        c.SeqCompute(
            [
                (EPD(ftrg) + 1, c.SetTo, nexttrg),
                (EPD(setter) + 4, c.SetTo, dst),
                (EPD(setter) + 5, c.SetTo, p),
            ]
        )
        c.SetNextTrigger(ftrg)
        nexttrg << c.NextTrigger()
        return ret

    raise EPError(_("Invalid input for EPD: {}").format(p))


# -------

_string_types = (bytes, str)


def FlattenList(lst: Any, ret=None) -> list:  # noqa: N802
    if ret is None:
        ret = []

    if isinstance(lst, _string_types) or hasattr(lst, "dont_flatten"):
        ret.append(lst)
        return ret

    try:
        for item in lst:
            FlattenList(item, ret)
        return ret

    except TypeError:  # lst is not iterable
        ret.append(lst)
        return ret


def FlattenIter(lst: Any) -> Iterator:  # noqa: N802
    if isinstance(lst, _string_types) or hasattr(lst, "dont_flatten"):
        yield lst
    else:
        try:
            for item in lst:
                yield from FlattenIter(item)
        except TypeError:
            yield lst


def List2Assignable(lst: Sequence[T]) -> T | Sequence[T]:  # noqa: N802
    if len(lst) == 1:
        return lst[0]

    else:
        return lst


@overload
def Assignable2List(a: None) -> list:
    ...


@overload
def Assignable2List(a: Iterable[T]) -> list[T]:
    ...


@overload
def Assignable2List(a: T) -> list[T]:
    ...


def Assignable2List(a: Any) -> list:  # noqa: N802
    if a is None:
        return []

    if not hasattr(a, "dont_flatten"):
        try:
            return list(a)
        except TypeError:
            pass

    return [a]


cachedfunc = functools.cache


# -------

# Original code from TrigEditPlus::ConvertString_SCMD2ToRaw


def SCMD2Text(s: str) -> str:  # noqa: N802
    #
    # normal -> xdigitinput1 -> xdigitinput2 -> xdigitinput3 -> normal
    #        '<'           xdigit          xdigit            '>'
    #                        -> normal
    #                       '>' emit '<>'
    #                                        -> normal
    #                                        '>' emit x00
    #                                                        -> normal
    # xdigit/normal  emit '<xx'
    def toxdigit(i: str) -> int | None:
        if "0" <= i <= "9":
            return ord(i) - 48
        elif "a" <= i <= "z":
            return ord(i) - 97 + 10
        elif "A" <= i <= "Z":
            return ord(i) - 65 + 10
        else:
            return None

    state: int = 0
    buf: list[int] = []
    bufch: list[str] = []
    out: list[str] = []

    # simple fsm
    for i in s:
        if state == 0:
            if i == "<":
                state = 1
                continue

            out.append(i)

        elif state == 1:
            xdi = toxdigit(i)
            if xdi is not None:
                buf.append(xdi)
                bufch.append(i)
                state = 2
                continue

            state = 0
            out.extend(["<", i])

        elif state == 2:
            xdi = toxdigit(i)
            if xdi is not None:
                buf.append(xdi)
                bufch.append(i)
                state = 3
                continue

            state = 0
            if i == ">":
                out.append(chr(buf.pop()))
                bufch.clear()

            else:
                out.extend(["<", bufch.pop(), i])
                buf.clear()

        elif state == 3:
            state = 0
            if i == ">":
                out.append(chr(buf[0] * 16 + buf[1]))

            else:
                out.extend(["<", bufch[0], bufch[1], i])
            buf.clear()
            bufch.clear()

    return "".join(out)


StrPath = str | os.PathLike[str]
BytesPath = bytes | os.PathLike[bytes]


@overload
def find_data_file(filename: StrPath, file: StrPath) -> str:
    ...


@overload
def find_data_file(filename: BytesPath, file: BytesPath) -> bytes:
    ...


def find_data_file(filename, file) -> str | bytes:
    if getattr(sys, "frozen", False):
        # The application is frozen
        datadir = os.path.dirname(sys.executable)
    else:
        # The application is not frozen
        # Change this bit to match where you store your data files:
        datadir = os.path.dirname(file)

    return os.path.join(datadir, filename)


def _rand_lst(lst: Iterable) -> list:
    lst = list(lst)
    random.shuffle(lst)
    return lst


class classproperty:  # noqa: N801
    """
    Decorator that converts a method with a single cls argument into a property
    that can be accessed directly from the class.
    """

    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)

    def getter(self, method):
        self.fget = method
        return self



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import ByteString, Collection
from typing import TypeAlias

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...trigtrg.runtrigtrg import _runner_cp

_shared_triggers: list[bytes | c.RawTrigger] = []
t_start_end: TypeAlias = tuple[c.ConstExpr, c.RawTrigger]


def get_executing_players(
    btrigger: bytes,
) -> list[bool]:
    # Get executing players of the trigger.
    # If AllPlayers executes it, then pass it
    if btrigger[320 + 2048 + 4 + 17] != 0:
        return [True] * 8
    # Should check manually, by player and force
    return [
        (
            bool(btrigger[320 + 2048 + 4 + player])
            or bool(btrigger[320 + 2048 + 4 + 18 + c.GetPlayerInfo(player).force])
        )
        for player in range(8)
    ]


def _preserved_and_has_no_wait(btrigger: ByteString) -> bool:
    exec_once = True
    for i in range(64):
        actionflag = btrigger[320 + 32 * i + 28]
        if not actionflag & 0b10:  # Enabled Flag
            continue
        actionbyte = btrigger[320 + 32 * i + 26]
        if actionbyte in (4, 7):  # Wait or Transmission
            return False
        elif actionbyte == 3:  # PreserveTrigger
            exec_once = False
    if exec_once and not (btrigger[320 + 2048] & 4):
        return False
    return True


def inline_codify_binary_trigger(btrigger: bytes) -> t_start_end:
    """Inline codify raw(binary) trigger data.

    For minimal protection, eudplib make some of the trig-triggers to
    eudplib trigger. This function makes eudplib trigger out of raw
    binary trigger stream.

    :param btrigger: Binary trigger data
    :returns: (tstart, tend) pair, as being used by tend
    """

    # 1. Get executing players of the trigger.
    # If all player executes it, then pass it
    executing_players = get_executing_players(btrigger)

    # 2. Create function body

    if c.PushTriggerScope():
        tstart = c.RawTrigger(actions=c.SetDeaths(0, c.SetTo, 0, 0))

        cp = _runner_cp
        execp = [p for p, e in enumerate(executing_players) if e]
        no_wait_and_preserved = _preserved_and_has_no_wait(btrigger)

        if no_wait_and_preserved and len(execp) == 8:
            tend = c.RawTrigger(trigSection=btrigger)
        if no_wait_and_preserved and max(execp) - min(execp) == len(execp) - 1:
            if cs.EUDIf()([cp >= min(execp), cp <= max(execp)]):
                c.RawTrigger(trigSection=btrigger)
            cs.EUDEndIf()
            tend = c.RawTrigger()
        else:
            cs.EUDSwitch(cp)
            for player in ut._rand_lst(range(8)):
                if executing_players[player]:
                    if cs.EUDSwitchCase()(player):
                        c.RawTrigger(trigSection=btrigger)
                        cs.EUDBreak()

            cs.EUDEndSwitch()

            tend = c.RawTrigger()
    c.PopTriggerScope()

    return tstart, tend


def _count_conditions_and_actions(btrigger: bytes) -> tuple[int, int]:
    cond_count, act_count = 0, 0
    for cond in range(16):
        if btrigger[cond * 20 + 15] == 22:  # Always
            continue
        elif btrigger[cond * 20 + 15] >= 1:
            cond_count += 1
    for act in range(64):
        if btrigger[320 + act * 32 + 26] in (47,):  # Comment
            continue
        elif btrigger[320 + act * 32 + 26] >= 1:
            act_count += 1
    return cond_count, act_count


def get_trigger_size(btrigger: bytes) -> int:
    cond_count, act_count = _count_conditions_and_actions(btrigger)
    min_size = 4 + 5 * cond_count + 8 * act_count
    trig = {
        "n": [1],  # nextptr
        "P": [594],  # preserved
        "C": list(range(2, 2 + 5 * cond_count)),  # conditions
        "c": [5 + 5 * cond_count],  # condtype == 0
        "A": list(range(82, 82 + 8 * act_count)),  # actions
        "a": [88 + 8 * act_count],  # acttype == 0
    }
    if cond_count == 16:
        del trig["c"]
    if act_count == 64:
        del trig["a"]

    for ret in range(min_size, 2408 // 4):
        output = ["_" for _ in range(ret)]

        def _f():
            for k, v in trig.items():
                for x in v:
                    if output[x % ret] != "_":
                        return False
                    output[x % ret] = k
            return True

        if _f():
            return ret * 4
    return 2408


def try_share_trigger(btrigger: bytes) -> int | bytes:
    if _preserved_and_has_no_wait(btrigger):
        _shared_triggers.append(btrigger)
        return len(_shared_triggers) - 1
    return btrigger


def inline_codify_binary_triggers(
    btriggers: Collection[bytes | int],
) -> t_start_end:
    """Inline codify raw(binary) trigger data.

    For minimal protection, eudplib make some of the trig-triggers to
    eudplib trigger. This function makes eudplib trigger out of raw
    binary trigger stream.

    :param btrigger: Binary trigger data
    :returns: (tstart, tend) pair, as being used by tend
    """

    # Create function body

    first_trigger = None
    t_startactions = [c.SetDeaths(0, c.SetTo, 0, 0)]
    b2s = False

    if c.PushTriggerScope():
        next_trigger = c.Forward()
        for i, btrigger in enumerate(btriggers):
            tend: c.RawTrigger
            if isinstance(btrigger, bytes):
                tend = c.RawTrigger(trigSection=btrigger)
                next_trigger << tend
                if i < len(btriggers) - 1:
                    next_trigger = c.Forward()
                b2s = True
            elif isinstance(btrigger, int):
                shared_trigger = _shared_triggers[btrigger]
                if isinstance(shared_trigger, bytes):
                    tend = c.RawTrigger(trigSection=shared_trigger)
                    _shared_triggers[btrigger] = tend
                elif isinstance(shared_trigger, c.RawTrigger):
                    tend = shared_trigger
                    if b2s:
                        c.SetNextTrigger(shared_trigger)
                else:
                    raise TypeError()
                next_trigger << tend
                if i < len(btriggers) - 1:
                    next_trigger = c.Forward()
                    t_startactions.append(c.SetNextPtr(tend, next_trigger))
                b2s = False
            else:
                raise TypeError()
            if first_trigger is None:
                first_trigger = tend
    c.PopTriggerScope()

    if first_trigger is None:
        raise TypeError()

    if c.PushTriggerScope():
        tstart = c.Forward()
        tstart << c.NextTrigger()

        cs.DoActions(t_startactions)

        c.SetNextTrigger(first_trigger)

    c.PopTriggerScope()

    return tstart, tend



from helper import *


@TestInstance
def test_ptrjump():
    testvar = EUDVariable()
    a = EUDVariable()
    t = Forward()

    testvar += 1
    a << t
    testvar += 2
    EUDJump(t)
    testvar += 4
    t << NextTrigger()
    testvar += 8

    PushTriggerScope()
    end = Forward()
    trg = [
        RawTrigger(nextptr=end, actions=testvar.AddNumber(16 << i)) for i in range(4)
    ]
    PopTriggerScope()
    EUDJump(trg[1])
    with expect_eperror():
        EUDJump(trg[2])
    with expect_eperror():
        EUDJump(trg[3])
    end << NextTrigger()

    test_assert("Pointer jump test", testvar == 43)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .eudarray import EUDArray
from .eudqueue import EUDDeque, EUDQueue
from .eudstack import EUDStack
from .playerv import PVariable
from .unitgroup import UnitGroup

__all__ = ["EUDArray", "EUDDeque", "EUDQueue", "EUDStack", "PVariable", "UnitGroup"]



from helper import *


@TestInstance
def test_unitloop():
    for ptr, epd in EUDLoopUnit2():
        pass


@TestInstance
def test_unitgroup():
    get_epd = lambda x: x if x._value._is_epd() else EPD(x)
    g = UnitGroup(1000)
    g.add(3)
    p, q = EUDVariable(2), EUDVariable(6)
    g.add(p)
    g.add(q)
    test_equality(
        "UnitGroup.add test", [p, q, f_dwread(p._varact + 24)], [2, 6, 0x072D0000]
    )

    k = EUDArray(7)
    for unit in g.cploop:  # 22, 21, 25
        if EUDIf()(False):
            for dead in unit.dying:
                pass
        EUDEndIf()
        # NOTE: Should we guarantee that first offset is always 0x4C?
        unit.move_cp(0x4C // 4)  # NO-OP
        f_dwadd_epd(EPD(0x6509B0), get_epd(k) - 19)
        f_dwwrite_cp(0, 1)
    test_equality(
        "Basic UnitGroup test", [k[i] for i in range(7)], [0, 0, 1, 1, 0, 0, 1]
    )

    g.add(4)
    g.add(EUDVariable(0))
    m = EUDArray(7 + 19)
    for unit in g.cploop:
        f_dwadd_epd(EPD(0x6509B0), get_epd(m))
        for dead in unit.dying:
            f_dwwrite_cp(0, 3)
            dead.move_cp(0)
            f_dwwrite_cp(0, 1)
        unit.move_cp(0)  # unit and dead are same object
        f_dwwrite_cp(0, 2)  # Never happen

    for unit in g.cploop:
        # NEVER run since all units were already removed from group 'g'
        f_dwadd_epd(EPD(0x6509B0), get_epd(m))
        f_dwwrite_cp(0, 2)
        unit.move_cp(0)
        f_dwwrite_cp(0, 2)
        unit.remove()  # this call is mandatory.
    indexes = (0, 2, 3, 4, 6)
    result = [
        1 if i in indexes else 3 if (i - 19) in indexes else 0 for i in range(26)
    ]
    test_equality("UnitGroup dying test", [m[i] for i in range(26)], result)

    count = EUDVariable()
    DoActions(CreateUnit(12, "Zerg Broodling", 64, P8), count.SetNumber(0))
    for cunit in EUDLoopPlayerCUnit(P8):
        count += 1
        Trigger(cunit.eqattr("unitType", "Zerg Broodling"), count.AddNumber(100))
    test_equality("EUDLoopPlayerCUnit test", count, 1212)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

from .. import core as c
from .. import trigger as tg
from .. import utils as ut
from ..core.variable import EUDLightBool, EUDVariable
from .basicstru import EUDJumpIf, EUDJumpIfNot


class EUDSCAnd:
    def __init__(self) -> None:
        self.const = True
        self.cond: list[c.Condition] = list()
        c.PushTriggerScope()
        self.side_effect = c.NextTrigger()
        self.scope = ut.EUDGetLastBlock()
        self.v: EUDLightBool | EUDVariable

    def patch(self) -> None:
        self.const = False
        self.jb = c.Forward()
        try:
            is_conditional = self.scope[1]["conditional"]
        except KeyError:
            is_conditional = False
        if is_conditional:
            self.v = EUDLightBool()
            self.fb = c.RawTrigger(nextptr=self.jb, actions=self.v.Clear())
            c.PopTriggerScope()
            c.RawTrigger(actions=self.v.Set())
        else:
            self.v = EUDVariable()
            self.fb = c.RawTrigger(nextptr=self.jb, actions=self.v.SetNumber(0))
            c.PopTriggerScope()
            self.v << True

    def __call__(
        self, cond=None, *, neg=False
    ) -> EUDSCAnd | list | EUDLightBool | EUDVariable:
        if cond is None:
            if self.const:
                c.PopTriggerScope()
                return self.cond
            elif self.cond:
                EUDJumpIfNot(self.cond, self.fb)
            self.jb << c.NextTrigger()
            return self.v

        if not self.const:
            if neg:
                EUDJumpIf(cond, self.fb)
            else:
                EUDJumpIfNot(cond, self.fb)
        else:
            is_const_cond = tg.tpatcher.is_const_cond(cond)
            nt_list = self.scope[1]["nexttrigger_list"]
            if nt_list != [self.side_effect]:
                # has side-effect
                ifcond = c.Forward()
                c.SetNextTrigger(ifcond)
                self.patch()
                if self.cond:
                    EUDJumpIfNot(self.cond, self.fb)
                    self.cond.clear()
                c.SetNextTrigger(self.side_effect)
                ifcond << c.NextTrigger()
                if neg:
                    EUDJumpIf(cond, self.fb)
                else:
                    EUDJumpIfNot(cond, self.fb)
            elif is_const_cond:
                if neg:
                    if tg.tpatcher.is_nagatable_cond(cond):
                        cond = tg.tpatcher.negate_cond(cond)
                        self.cond.append(cond)
                    else:
                        self.patch()
                        if self.cond:
                            EUDJumpIfNot(self.cond, self.fb)
                            self.cond.clear()
                        EUDJumpIf(cond, self.fb)
                else:
                    self.cond.append(cond)
            else:
                self.patch()
                # TODO: handle mixing of non/side-effect conditions
                if self.cond:
                    EUDJumpIfNot(self.cond, self.fb)
                    self.cond.clear()
                if neg:
                    EUDJumpIf(cond, self.fb)
                else:
                    EUDJumpIfNot(cond, self.fb)
        return self


class EUDSCOr:
    def __init__(self) -> None:
        self.jb = c.Forward()
        self.v = EUDLightBool()
        c.RawTrigger(actions=self.v.Clear())

        if c.PushTriggerScope():
            self.tb = c.RawTrigger(nextptr=self.jb, actions=self.v.Set())
        c.PopTriggerScope()

    def __call__(self, cond=None, *, neg=False) -> EUDSCOr | EUDLightBool:
        if cond is None:
            self.jb << c.NextTrigger()
            return self.v

        else:
            if neg:
                EUDJumpIfNot(cond, self.tb)
            else:
                EUDJumpIf(cond, self.tb)
            return self



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import string

from .. import core as c
from .. import utils as ut
from ..localize import _
from .cpprint import PColor, PName, eprint_all, f_cpstr_print, f_eprintln
from .dbstr import DBString
from .eudprint import f_dbstr_print
from .strcommon import epd2s, hptr, ptr2s


class _EUDFormatter(string.Formatter):
    def __init__(self):
        self.last_number = 0

    def vformat(self, format_string, args, kwargs):
        used_args = set()
        try:
            result, _i = self._eudformat(format_string, args, kwargs, used_args, 2)
        except IndexError:
            raise ut.EPError(_("Too many placeholders in the format string."))
        self.check_unused_args(used_args, args, kwargs)
        return result

    def _eudformat(
        self,
        format_string,
        args,
        kwargs,
        used_args,
        recursion_depth,
        auto_arg_index=0,
    ):
        if recursion_depth < 0:
            raise ValueError(_("Max string recursion exceeded"))
        result = []
        for literal_text, field_name, format_spec, conversion in self.parse(
            format_string
        ):
            # output the literal text
            if literal_text:
                result.append(literal_text)

            # if there's a field, output it
            if field_name is not None:
                # this is some markup, find the object and do the formatting

                # fmt: off
                # handle arg indexing when empty field_names are given.
                if field_name == "":
                    if auto_arg_index is False:
                        raise ValueError(_("cannot switch from manual field specification to automatic field numbering"))  # noqa: E501
                    field_name = str(auto_arg_index)
                    auto_arg_index += 1
                elif field_name.isdigit():
                    if auto_arg_index:
                        raise ValueError(_("cannot switch from manual field specification to automatic field numbering"))  # noqa: E501
                    # disable auto arg incrementing, if it gets
                    # used later on, then an exception will be raised
                    auto_arg_index = False

                # fmt: on
                # given the field_name, find the object it references
                #  and the argument it came from
                obj, arg_used = self.get_field(field_name, args, kwargs)
                used_args.add(arg_used)

                # do any conversion on the resulting object
                obj = self.convert_field(obj, conversion)

                # expand the format spec, if needed
                format_spec, auto_arg_index = self._vformat(
                    format_spec,
                    args,
                    kwargs,
                    used_args,
                    recursion_depth - 1,
                    auto_arg_index=auto_arg_index,
                )

                # format the object and append to the result
                result.append(self.eudformat_field(obj, format_spec))

        return result, auto_arg_index

    def eudformat_field(self, value, format_spec):
        if format_spec.endswith("s"):
            if c.IsConstExpr(value):
                try:
                    if value % 4 == 0:
                        return epd2s(ut.EPD(value))
                except TypeError:
                    return ptr2s(value)
            return ptr2s(value)
        elif format_spec.endswith("t"):
            return epd2s(value)
        elif format_spec.endswith("x") or format_spec.endswith("X"):
            return hptr(value)
        elif format_spec.endswith("c"):
            return PColor(value)
        elif format_spec.endswith("n"):
            return PName(value)
        else:
            types = (c.Db, ptr2s, epd2s, hptr, c.EUDVariable, DBString)
            for t in types:
                if ut.isUnproxyInstance(value, t):
                    return value
            if c.IsConstExpr(value):
                return value
        return format(value, format_spec)

    def get_value(self, key, args, kwargs):
        if key == "":
            key = self.last_number
            self.last_number += 1
        return super().get_value(key, args, kwargs)


def _format_args(format_string, *args):
    formatter = _EUDFormatter()
    return formatter.format(format_string, *args)


def f_sprintf(dst, format_string, *args, EOS=True, encoding="UTF-8"):  # noqa: N803
    fmtargs = _format_args(format_string, *args)
    return f_dbstr_print(dst, *fmtargs, EOS=EOS, encoding=encoding)


def f_sprintf_cp(format_string, *args, EOS=True, encoding="UTF-8"):  # noqa: N803
    fmtargs = _format_args(format_string, *args)
    f_cpstr_print(*fmtargs, EOS=EOS, encoding=encoding)


def f_eprintf(format_string, *args):
    fmtargs = _format_args(format_string, *args)
    f_eprintln(*fmtargs)


def f_eprintAll(format_string, *args):  # noqa: N802
    fmtargs = _format_args(format_string, *args)
    eprint_all(*fmtargs)



from helper import *

cunitepd_old = f_readgen_epd(
    0x7FFFF8, (0, lambda x: x), (-0x58A364 // 4, lambda y: y // 4)
)


@TestInstance
def test_perfmemio():
    t = {}
    for i in range(1700):
        x = 0x59CCA8 + 336 * i
        try:
            t[bin(x).count("1")] = x
        except KeyError:
            continue
    c, ptr = EUDCreateVariables(2)
    v = EPD(c.getValueAddr())
    ptr << 0x58A364
    for i in range(5, 18 + 1):
        c << t[i]
        test_perf("cunitepd_old(bit%u)" % i, lambda: cunitepd_old(v), perf_basecount)
        test_perf(
            "epdcunit_new(bit%u)" % i, lambda: f_cunitepdread_epd(v), perf_basecount
        )
    c << 0
    test_perf("f_dwread_epd", lambda: f_dwread_epd(c), perf_basecount)
    test_perf("f_dwbreak", lambda: f_dwbreak(ptr), perf_basecount)
    test_perf("EPD", lambda: EPD(ptr), perf_basecount)
    test_perf("f_dwread", lambda: f_dwread(ptr), perf_basecount // 2)
    test_perf("f_wread", lambda: f_wread(ptr), perf_basecount // 2)
    test_perf("f_bread", lambda: f_bread(ptr), perf_basecount // 2)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import random

from ... import core as c
from ... import ctrlstru as cs
from ... import memio as mi
from ... import utils as ut
from ...core.mapdata.chktok import CHK
from ...eudlib.utilf.gametick import f_getgametick
from ...scdata import CurrentPlayer
from ...trigtrg import runtrigtrg as rtt
from ..inlinecode.ilcprocesstrig import _get_inline_code_list

""" Stage 3:
- Fixes nextptr modification to TRIG triggers by stage 1
- Flip property value by trig section
- Dispatch inline codes
- Create infinite executer for root
"""


def _dispatch_inline_code(
    nextptr: c.EUDVariable, trigepd: c.EUDVariable, prop: int
) -> None:
    cs0 = c.Forward()  # set cs0+20 to code_start
    cs1 = c.Forward()  # set cs1+20 to ut.EPD(code_end) + 1
    cs2 = c.Forward()  # set cs2+20 to cs_a0_epd + 4
    cs3 = c.Forward()  # set cs3+20 to cs_a0_epd + 5
    resetter = c.Forward()
    end = c.Forward()

    v = random.randint(234, 65535)
    c.VProc(
        trigepd,
        [
            trigepd.AddNumber(-prop + 2 - 12 * v),
            trigepd.SetDest(ut.EPD(0x6509B0)),
        ],
    )

    for func_id, func in _get_inline_code_list():
        code_start, code_end = func
        cs_a0_epd = ut.EPD(code_start) + (8 + 320) // 4

        t = c.Forward()
        nt = c.Forward()
        acts = [
            c.SetMemory(cs0 + 20, c.SetTo, code_start),
            c.SetMemory(cs1 + 20, c.SetTo, ut.EPD(code_end + 4)),
            c.SetMemory(cs2 + 20, c.SetTo, cs_a0_epd + 4),
            c.SetMemory(cs3 + 20, c.SetTo, cs_a0_epd + 5),
            c.SetMemory(resetter + 16, c.SetTo, ut.EPD(t + 4)),
            c.SetMemory(resetter + 20, c.SetTo, nt),
            c.SetNextPtr(t, end),
        ]
        t << c.RawTrigger(
            conditions=[c.Deaths(CurrentPlayer, c.Exactly, func_id, v)],
            actions=ut._rand_lst(acts),
        )
        nt << c.NextTrigger()

    end << c.NextTrigger()

    st1, st2, st3, st4, stf = (c.Forward() for _ in range(5))
    c.VProc(
        [trigepd, nextptr],
        [
            trigepd.AddNumber(12 * v - 1),
            # Reset t's nextptr
            resetter << c.SetNextPtr(0, 0),
            # SetNextPtr for this trigger
            # action #1
            c.SetMemory(0x6509B0, c.Add, -2 + (8 + 320) // 4 + 4),
            trigepd.SetDest(ut.EPD(st1) + 5),
            nextptr.SetDest(ut.EPD(st2) + 5),
        ],
    )
    c.RawTrigger(
        nextptr=trigepd.GetVTable(),
        actions=[
            st1 << c.SetDeaths(CurrentPlayer, c.SetTo, 0, v),
            c.SetMemory(0x6509B0, c.Add, 1),
            cs0 << c.SetDeaths(CurrentPlayer, c.SetTo, 0, v),
            # SetNextPtr for codeend
            # action #2
            c.SetMemory(0x6509B0, c.Add, 7),
            cs1 << c.SetDeaths(CurrentPlayer, c.SetTo, 0, v),
            c.SetMemory(0x6509B0, c.Add, 1),
            st2 << c.SetDeaths(CurrentPlayer, c.SetTo, 0, v),
            # This trigger sets argument for cs_a0_epd
            # cs_a0 will be SetNextPtr(trigepd + 1, nextptr) after this
            # action #3
            c.SetMemory(0x6509B0, c.Add, 7),
            cs2 << c.SetDeaths(CurrentPlayer, c.SetTo, 0, v),
            c.SetMemory(0x6509B0, c.Add, 1),
            trigepd.SetDest(ut.EPD(st3) + 5),
            nextptr.SetDest(ut.EPD(st4) + 5),
            c.SetNextPtr(nextptr.GetVTable(), stf),
        ],
    )
    stf << c.RawTrigger(
        actions=[
            st3 << c.SetDeaths(CurrentPlayer, c.SetTo, 0, v),
            # action #4
            c.SetMemory(0x6509B0, c.Add, 7),
            cs3 << c.SetDeaths(CurrentPlayer, c.SetTo, 0, v),
            c.SetMemory(0x6509B0, c.Add, 1),
            st4 << c.SetDeaths(CurrentPlayer, c.SetTo, 0, v),
        ]
    )


def _flip_prop(trigepd: c.EUDVariable) -> None:
    """Iterate through triggers and flip 'Trigger disabled' flag

    Also, dispatch inline codes
    """

    if cs.EUDWhileNot()([0x3FD56E6E <= trigepd, trigepd <= 0x3FD56E86]):
        # trigepd's nextptr may change during flipping process, so
        # we get it now.
        trigepd += 1
        nexttrig, nexttrigepd = mi.f_dwepdread_epd(trigepd)
        u = random.randint(234, 65535)
        prop = (8 + 320 + 2048) // 4 - 12 * u
        c.VProc(
            trigepd,
            [
                trigepd.AddNumber(prop - 1),
                trigepd.SetDest(ut.EPD(0x6509B0)),
            ],
        )

        # Toggle Preserved Flag
        is_inline_code, start_dispatch, set_trigepd = (c.Forward() for _ in range(3))
        c.RawTrigger(
            actions=[
                c.SetDeathsX(CurrentPlayer, c.Add, 8, u, 8),
                c.SetNextPtr(is_inline_code, set_trigepd),
            ]
        )

        # Dispatch inline code
        is_inline_code << c.RawTrigger(
            nextptr=set_trigepd,
            conditions=c.Deaths(CurrentPlayer, c.Exactly, 0x10000000, u),
            actions=[
                c.SetNextPtr(is_inline_code, start_dispatch),
                c.SetDeaths(CurrentPlayer, c.SetTo, 4, u),  # Preserve
            ],
        )
        start_dispatch << c.NextTrigger()
        _dispatch_inline_code(nexttrig, trigepd, prop)

        set_trigepd << c.NextTrigger()
        trigepd << nexttrigepd

    cs.EUDEndWhile()


def create_inject_finalizer(
    chkt: CHK, root: c.Forward | c.RawTrigger, mrgndata: bytes | None = None
) -> c.Forward:
    rtt._alloc_trigtrigger_link()
    c.EP_SetRValueStrictMode(False)

    pts = 0x51A280
    mrgn = 0x58DC60

    # Apply inline code patch
    if c.PushTriggerScope():
        ret = c.NextTrigger()

        # Revert nextptr
        triggerend = mi.f_dwread_epd(9)
        ptsprev_epd = mi.f_dwread_epd(10)
        c.VProc(
            [ptsprev_epd, triggerend],
            [
                c.SetDeaths(9, c.SetTo, 0, 0),
                c.SetDeaths(10, c.SetTo, 0, 0),
                ptsprev_epd.SetDest(ut.EPD(triggerend.getDestAddr())),
            ],
        )

        c.PushTriggerScope()
        mi.f_repmovsd_epd(0, 0, 0)
        c.PopTriggerScope()

        # revert mrgndata
        if mrgndata is None:
            mrgndata = chkt.getsection("MRGN")[: 2408 + 836]
            mrgndata_db = c.Db(mrgndata)
            mi.f_repmovsd_epd(ut.EPD(mrgn), ut.EPD(mrgndata_db), len(mrgndata) // 4)
        else:
            from ...memio.mblockio import _copydwn, _read_end

            mrgndata_db = c.Db(mrgndata)
            nexttrg = c.Forward()
            c.RawTrigger(
                nextptr=mi.f_repmovsd_epd._fstart,
                actions=[
                    c.SetMemory(_read_end + 344, c.SetTo, ut.EPD(mrgn)),
                    c.SetMemoryX(_read_end + 352, c.SetTo, 8 << 24, 0xFF000000),
                    c.SetMemory(0x6509B0, c.SetTo, ut.EPD(mrgndata_db)),
                    c.SetMemory(_copydwn, c.SetTo, len(mrgndata) // 4),
                    c.SetMemory(mi.f_repmovsd_epd._nptr, c.SetTo, nexttrg),
                ],
            )
            nexttrg << c.RawTrigger(
                actions=c.SetMemoryX(_read_end + 352, c.SetTo, 7 << 24, 0xFF000000)
            )

        # Flip TRIG properties
        i = c.EUDVariable(ut.EPD(pts + 8))
        if cs.EUDWhileNot()(i >= ut.EPD(pts + 8) + 3 * 7 + 1):
            _flip_prop(mi.f_epdread_epd(i))
            i += 3
        cs.EUDEndWhile()

        # Create payload for each players & Link them with pts
        lasttime = c.EUDVariable()
        tmcheckt = c.Forward()

        for player in ut._rand_lst(range(8)):
            trs = rtt._runner_start[player]
            tre = rtt._runner_end[player]

            c.PushTriggerScope()

            # Crash preventer
            tstart, _t0 = c.Forward(), c.Forward()
            tstart << c.RawTrigger(
                prevptr=pts + player * 12 + 4,
                nextptr=trs,
                actions=c.SetNextPtr(tstart, _t0),
            )

            _t0 << c.RawTrigger(
                nextptr=tmcheckt, actions=c.SetNextPtr(tstart, trs)
            )  # reset

            c.PopTriggerScope()

            prevtstart = mi.f_dwread_epd(ut.EPD(pts) + player * 3 + 2)
            prevtend, prevtend_epd = mi.f_dwepdread_epd(ut.EPD(pts) + player * 3 + 1)

            # If there were triggers
            if cs.EUDIfNot()(prevtstart == ~(pts + player * 12 + 4)):
                (
                    orig_tstart,
                    orig_tend,
                    _runner_end_array,
                ) = rtt._alloc_trigtrigger_link()
                link_trs = c.Forward()
                vs = [prevtstart, prevtend, prevtend_epd]
                acts = [
                    # Link pts
                    c.SetMemory(pts + player * 12 + 8, c.SetTo, tstart),
                    c.SetMemory(pts + player * 12 + 4, c.SetTo, tre),
                    # Cache dlist start & end
                    prevtstart.SetDest(
                        orig_tstart + player
                        if orig_tstart._is_epd()
                        else ut.EPD(orig_tstart) + player
                    ),
                    prevtend.SetDest(
                        orig_tend + player
                        if orig_tend._is_epd()
                        else ut.EPD(orig_tend) + player
                    ),
                    prevtend_epd.AddNumber(1),
                    prevtend_epd.SetDest(ut.EPD(link_trs) + 4),
                ]
                c.VProc(ut._rand_lst(vs), ut._rand_lst(acts))
                c.VProc(
                    prevtstart,
                    [
                        # Link trs
                        link_trs << c.SetDeaths(0, c.SetTo, tre, 0),
                        prevtstart.SetDest(ut.EPD(trs) + 1),
                    ],
                )
            cs.EUDEndIf()

        if c.PushTriggerScope():
            tmcheckt << c.NextTrigger()
            if cs.EUDIfNot()(c.Memory(0x57F23C, c.Exactly, lasttime)):
                f_getgametick(ret=[lasttime])
                c.SetNextTrigger(root)
            cs.EUDEndIf()

            # Set current player to 1.
            c.RawTrigger(
                nextptr=0x80000000, actions=c.SetMemory(0x6509B0, c.SetTo, 0)
            )
        c.PopTriggerScope()

        # update lasttime
        f_getgametick(ret=[lasttime])
        c.RawTrigger(actions=c.SetMemory(0x6509B0, c.SetTo, 0))  # Current player = 1

        # now jump to root
        c.SetNextTrigger(root)

    c.PopTriggerScope()

    return ret



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .bytedump import Db
from .eudobj import EUDObject

__all__ = ["Db", "EUDObject"]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import functools

from ... import utils as ut
from ...localize import _
from ...utils.blockstru import BlockStruManager, set_blockstru_manager
from .. import allocator as ac
from .. import rawtrigger as bt
from .. import variable as ev
from .trace.tracetool import _eud_trace_pop, _eud_trace_push

_current_compiled_func = None
_current_trigger_count = 0


def _update_func_trigger_count():
    global _current_trigger_count
    current_counter = bt.GetTriggerCounter()
    added_trigger_count = current_counter - _current_trigger_count

    if _current_compiled_func:
        _current_compiled_func._triggerCount += added_trigger_count
    _current_trigger_count = current_counter


def _set_current_compiled_func(func):
    global _current_compiled_func

    last_compiled_func = _current_compiled_func
    _update_func_trigger_count()
    _current_compiled_func = func
    return last_compiled_func


class EUDFuncN:
    def __init__(self, argn, callerfunc, bodyfunc, *, traced):
        """EUDFuncN

        :param callerfunc: Function to be wrapped.
        :param argn: The number of arguments got by callerfunc
        :param bodyfunc: Where function should return to
        """

        if bodyfunc is None:
            bodyfunc = callerfunc

        if isinstance(argn, int):
            self._argn = argn
            self._arginits = None
        elif all(
            isinstance(initvals, tuple) and 3 <= len(initvals) <= 4
            for initvals in argn
        ):
            self._argn = len(argn)
            self._arginits = argn
        else:
            raise ut.EPError(f"Invalid argn: {argn}")
        self._retn = None
        self._callerfunc = callerfunc
        self._bodyfunc = bodyfunc
        functools.update_wrapper(self, bodyfunc)
        self._fstart = None
        self._fend = None
        self._nptr = None
        self._fargs = None
        self._frets = None
        self._triggerCount = None
        self._traced = traced

    def size(self):
        if not self._fstart:
            self._create_func_body()

        return self._triggerCount

    def _create_func_body(self):
        self._triggerCount = 0
        last_compiled_func = _set_current_compiled_func(self)

        # Add return point
        self._fend = ac.Forward()

        # Prevent double compilication
        ut.ep_assert(self._fstart is None, "EUDFuncN was already compiled")

        # Initialize new namespace
        f_bsm = BlockStruManager()
        prev_bsm = set_blockstru_manager(f_bsm)
        bt.PushTriggerScope()

        self._create_func_args()

        fstart = bt.NextTrigger()
        self._fstart = fstart

        if self._traced:
            _eud_trace_push()

        final_rets = self._callerfunc(*self._fargs)
        if final_rets is not None:
            self._add_return(ut.Assignable2List(final_rets), False)

        if not self._fend.IsSet():
            self._fend << bt.NextTrigger()
        if self._traced:
            _eud_trace_pop()
        if self._retn is None or self._retn == 0:
            self._fend = bt.RawTrigger()
            self._nptr = self._fend + 4
        else:
            fend_trgs = ut.FlattenList(ev.VProc(self._frets, []))
            self._fend = fend_trgs[0]
            self._nptr = fend_trgs[-1]._actions[-1] + 20

        bt.PopTriggerScope()

        # Finalize
        if not f_bsm.empty():
            raise ut.EPError(
                _("Block start/end mismatch inside function")
                + f": {', '.join(name for name, _data in f_bsm._blockstru)}"
            )
        set_blockstru_manager(prev_bsm)

        # No return -> set return count to 0
        if self._retn is None:
            self._retn = 0
        _set_current_compiled_func(last_compiled_func)

    def _create_func_args(self):
        if self._fargs is None:
            if self._arginits is None:
                self._fargs = [ev.EUDVariable() for _ in range(self._argn)]
            else:
                self._fargs = []
                for initvals in self._arginits:
                    argv = ev.EUDXVariable(*initvals)
                    self._fargs.append(argv)

    def _add_return(self, retv, needjump):
        retv = ut.FlattenList(retv)
        if self._frets is None:
            self._frets = [ev.EUDVariable() for _ in retv]
            self._retn = len(retv)

        ut.ep_assert(
            len(retv) == len(self._frets),
            _("Number of returned values should be constant.")
            + _(" (From function %s)").format(self._bodyfunc.__name__),
        )

        var_assigns, const_assigns = list(), list()
        for fret, ret in zip(self._frets, retv):
            if ev.IsEUDVariable(ret):
                var_assigns.append((fret, bt.SetTo, ret))
            else:
                const_assigns.append((fret, bt.SetTo, ret))
        if len(var_assigns) <= 2:
            ev.SeqCompute(const_assigns + var_assigns)
        else:
            ev.NonSeqCompute(const_assigns + var_assigns)

        if needjump:
            bt.SetNextTrigger(self._fend)

    def _call_with_last_args(self, *, ret=None):
        if self._fstart is None:
            self._create_func_body()

        fcallend = ac.Forward()

        # SeqCompute gets faster when const-assignments are in front of
        # variable assignments.
        if ret is None:
            ret = ev.EUDCreateVariables(self._retn)
        ret = ut.FlattenList(ret)
        ut.ep_assert(
            len(ret) == self._retn,
            _("Return number mismatch : ")
            + "len(%s) != %d" % (repr(ret), self._retn),
        )
        next_ptr_assignment = [(ut.EPD(self._nptr), bt.SetTo, fcallend)]
        if self._retn >= 1:
            for fret, retv in zip(self._frets, ret):
                try:
                    retv = ut.EPD(retv.getValueAddr())
                except AttributeError:
                    pass
                next_ptr_assignment.append(
                    (ut.EPD(fret.getDestAddr()), bt.SetTo, retv)
                )

        ev.SeqCompute(next_ptr_assignment)
        bt.SetNextTrigger(self._fstart)

        fcallend << bt.NextTrigger()

        if self._frets is not None:
            for retv in ret:
                try:
                    retv.makeR()
                except AttributeError:
                    pass
            return ut.List2Assignable(ret)

    def __call__(self, *args, ret=None):
        if self._fstart is None:
            self._create_func_body()

        ut.ep_assert(
            len(args) == self._argn,
            _("Argument number mismatch : ")
            + "len(%s) != %d" % (repr(args), self._argn),
        )

        fcallend = ac.Forward()

        # SeqCompute gets faster when const-assignments are in front of
        # variable assignments.
        if ret is None:
            ret = ev.EUDCreateVariables(self._retn)
        ret = ut.FlattenList(ret)
        ut.ep_assert(
            len(ret) == self._retn,
            _("Return number mismatch : ")
            + "len(%s) != %d" % (repr(ret), self._retn),
        )
        next_ptr_assignment = [(ut.EPD(self._nptr), bt.SetTo, fcallend)]
        if self._retn >= 1:
            for fret, retv in zip(self._frets, ret):
                try:
                    retv = ut.EPD(retv.getValueAddr())
                except AttributeError:
                    pass
                next_ptr_assignment.append(
                    (ut.EPD(fret.getDestAddr()), bt.SetTo, retv)
                )
        var_assigns, const_assigns = list(), list()
        for farg, arg in zip(self._fargs, args):
            if ev.IsEUDVariable(arg):
                var_assigns.append((farg, bt.SetTo, arg))
            else:
                const_assigns.append((farg, bt.SetTo, arg))
        if len(var_assigns) <= 2:
            ev.SeqCompute(next_ptr_assignment + const_assigns + var_assigns)
        else:
            ev.NonSeqCompute(next_ptr_assignment + const_assigns + var_assigns)
        bt.SetNextTrigger(self._fstart)

        fcallend << bt.NextTrigger()

        if self._frets is not None:
            for retv in ret:
                try:
                    retv.makeR()
                except AttributeError:
                    pass
            return ut.List2Assignable(ret)


def EUDReturn(*args):  # noqa: N802
    _current_compiled_func._add_return(args, True)
